import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from uuid import UUID

import wrapt  # type: ignore

from layer import Dataset, Model
from layer.clients.layer import LayerClient
from layer.config import ConfigManager
from layer.config.config import Config
from layer.context import Context
from layer.contracts.assertions import Assertion
from layer.contracts.assets import AssetType
from layer.contracts.datasets import DatasetBuild, DatasetBuildStatus
from layer.contracts.runs import FunctionDefinition
from layer.contracts.tracker import DatasetTransferState
from layer.decorators.layer_wrapper import LayerAssetFunctionWrapper
from layer.global_context import reset_active_context, set_active_context
from layer.projects.project_runner import register_dataset_function
from layer.projects.utils import (
    get_current_project_full_name,
    verify_project_exists_and_retrieve_project_id,
)
from layer.settings import LayerSettings
from layer.tracker.progress_tracker import RunProgressTracker
from layer.utils.async_utils import asyncio_run_in_thread
from layer.utils.runtime_utils import check_and_convert_to_df


logger = logging.getLogger(__name__)


def dataset(
    name: str, dependencies: Optional[List[Union[str, Dataset, Model]]] = None
) -> Callable[..., Any]:
    """
    Decorator function that wraps a dataset function.

    The decorator ensures that:
    - If decorated function is passed in to ``layer.run(your_function)``, then Layer runs the function remotely and stores its output as a dataset.
    - If you run the function locally, ``your_function()``, then Layer stores the output in the Layer backend as a dataset. This does not affect function execution.
    - The result of the function is a pandas.DataFrame or an array of column name -> value dict-s, which under the hood is converted to a pandas.DataFrame

    Supported Pandas dataframe types:
    * `bool`
    * `(u)int{8,16,32,64}`
    * `float32`
    * `float64`
    * `str`
    * `pd.Categorical` (categorical maps to only integer categories, label information is lost during the conversion)
    * `pd.Timestamp`
    * `datetime.date`
    * `datetime.time`
    * `PIL.Image.Image`
    * `numpy.ndarray`

    :param name: Name with which the dataset will be stored in Layer backend.
    :param dependencies: List of ``Datasets`` or ``Models`` that will be built by Layer backend prior to building the current function. This hints Layer what entities this function depends on and optimizes the build process.
    :return: Function object being decorated.

    .. code-block:: python

        import pandas as pd
        from layer
        from layer.decorators import dataset

        # Define a new function for dataset generation:
        # - The dependencies list includes entities that needs to be built before running the `create_my_dataset` code
        # - `titanic` is a publicly accessible dataset to everyone using Layer
        @dataset("my_titanic_dataset", dependencies=Dataset("titanic"))
        def create_my_titanic_dataset():
            df = layer.get_dataset("titanic").to_pandas()
            return df

        # Run function locally
        df = create_my_titanic_dataset()
        # Dataset will be stored in Layer backend and will be retrievable later
        assert df == layer.get_dataset("my_titanic_dataset").to_pandas()

    Here's another way to create a dataset.
    As long as the function outputs a Pandas dataframe or an array of dict-s, Layer doesn't really care how you load the data.

    .. code-block:: python

        import pandas as pd
        from layer.decorators import dataset

        @dataset("my_products")
        def create_product_dataset():
            data_as_dict = [
                {"Id": 1, "Product": "product1", "Price": 15},
                {"Id": 2, "Product": "product2", "Price": 20},
                {"Id": 3, "Product": "product3", "Price": 10}
            ]
            data = [[1, "product1", 15], [2, "product2", 20], [3, "product3", 10]]
            return pd.DataFrame(data, columns=["Id", "Product", "Price"])
            #  return data_as_dict  # You can also return this

        product_dataset = create_product_dataset()


    If you use a dataset that is outside of your project, then you must explicitly call it out as a dependency.
    These dependencies are displayed in the **External assets** section in your Layer project.

    .. code-block:: python

        import pandas as pd
        from layer
        from layer.decorators import dataset

        @dataset("raw_spam_dataset", dependencies=[layer.Dataset('layer/spam-detection/datasets/spam_messages')])
        def raw_spam_dataset():
            # Get the spam_messages dataset and convert to Pandas dataframe.
            df = layer.get_dataset("spam-detection/datasets/spam_messages").to_pandas()
            return df
    """

    @wrapt.decorator(proxy=_dataset_wrapper(name, dependencies))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _dataset_wrapper(
    name: str, dependencies: Optional[List[Union[str, Dataset, Model]]] = None
) -> Any:
    class DatasetFunctionWrapper(LayerAssetFunctionWrapper):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any) -> None:
            super().__init__(
                wrapped,
                wrapper,
                enabled,
                AssetType.DATASET,
                name,
                dependencies,
            )

        # This is not serialized with cloudpickle, so it will only be run locally.
        # See https://layerco.slack.com/archives/C02R5B3R3GU/p1646144705414089 for detail.
        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            self.layer.validate()
            current_project_full_name_ = get_current_project_full_name()
            config: Config = asyncio_run_in_thread(ConfigManager().refresh())
            with LayerClient(config.client, logger).init() as client:
                progress_tracker = RunProgressTracker(
                    url=config.url,
                    project_name=current_project_full_name_.project_name,
                    account_name=current_project_full_name_.account_name,
                )
                with progress_tracker.track() as tracker:
                    tracker.add_asset(AssetType.DATASET, self.layer.get_asset_name())
                    dataset_definition = self.get_definition()
                    result = _build_dataset_locally_and_store_remotely(
                        lambda: super(  # pylint: disable=super-with-arguments
                            DatasetFunctionWrapper, self
                        ).__call__(*args, **kwargs),
                        self.layer,
                        dataset_definition,
                        tracker,
                        client,
                    )
                    return result

    return DatasetFunctionWrapper


def _build_dataset_locally_and_store_remotely(
    building_func: Callable[..., Any],
    layer: LayerSettings,
    dataset: FunctionDefinition,
    tracker: RunProgressTracker,
    client: LayerClient,
) -> Any:
    tracker.add_asset(AssetType.DATASET, layer.get_asset_name())

    register_dataset_function(client, dataset, True, tracker)
    tracker.mark_dataset_building(layer.get_asset_name())

    (result, build_uuid) = _build_locally_update_remotely(
        client,
        building_func,
        dataset,
        tracker,
        layer.get_assertions(),
    )

    transfer_state = DatasetTransferState(len(result))
    tracker.mark_dataset_saving_result(dataset.asset_name, transfer_state)

    # this call would store the resulting dataset, extract the schema and complete the build from remote
    client.data_catalog.store_dataset(
        name="",
        data=result,
        build_id=build_uuid,
        progress_callback=transfer_state.increment_num_transferred_rows,
    )
    tracker.mark_dataset_built(dataset.asset_name)
    return result


def _build_locally_update_remotely(
    client: LayerClient,
    function_that_builds_dataset: Callable[..., Any],
    dataset: FunctionDefinition,
    tracker: RunProgressTracker,
    assertions: List[Assertion],
) -> Tuple[Any, UUID]:
    try:
        with Context() as context:
            # TODO pass path to API instead
            current_project_uuid = verify_project_exists_and_retrieve_project_id(
                client, dataset.project_full_name
            )
            initiate_build_response = client.data_catalog.initiate_build(
                dataset,
                current_project_uuid,
                True,
            )
            dataset_build_id = UUID(initiate_build_response.id.value)
            context.with_dataset_build(
                DatasetBuild(id=dataset_build_id, status=DatasetBuildStatus.STARTED)
            )
            context.with_tracker(tracker)
            context.with_asset_name(dataset.asset_name)
            set_active_context(context)
            try:
                result = function_that_builds_dataset()
                result = check_and_convert_to_df(result)
                _run_assertions(dataset.asset_name, result, assertions, tracker)
            except Exception as e:
                client.data_catalog.complete_build(
                    initiate_build_response.id, dataset, e
                )
                context.with_dataset_build(
                    DatasetBuild(id=dataset_build_id, status=DatasetBuildStatus.FAILED)
                )
                raise e
            reset_active_context()

            context.with_dataset_build(
                DatasetBuild(id=dataset_build_id, status=DatasetBuildStatus.COMPLETED)
            )
            return result, UUID(str(initiate_build_response.id.value))
    finally:
        reset_active_context()


def _run_assertions(
    asset_name: str,
    result: Any,
    assertions: List[Assertion],
    tracker: RunProgressTracker,
) -> None:
    failed_assertions = []
    tracker.mark_dataset_running_assertions(asset_name)
    for assertion in reversed(assertions):
        try:
            tracker.mark_dataset_running_assertion(asset_name, assertion)
            assertion.function(result)
        except Exception:
            failed_assertions.append(assertion)
    if len(failed_assertions) > 0:
        tracker.mark_dataset_failed_assertions(asset_name, failed_assertions)
        raise Exception(f"Failed assertions {failed_assertions}\n")
    else:
        tracker.mark_dataset_completed_assertions(asset_name)



import re
import unicodedata


def slugify(value: str, allow_unicode: bool = False) -> str:
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.

    Taken unchanged from django/utils/text.py
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value).strip().lower()
    return re.sub(r"[-\s]+", "-", value)



import os
import tarfile
from pathlib import Path


def tar_directory(output_filename: str, source_dir: Path) -> None:
    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname=os.path.sep)



from pathlib import Path

import pandas as pd
from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PbModelFlavor

from layer.types import ModelObject

from .base import ModelFlavor, ModelRuntimeObjects


class HuggingFaceModelFlavor(ModelFlavor):
    """An ML Model flavor implementation which handles persistence of Hugging Face Transformer Models."""

    MODULE_KEYWORD = "transformers.models"
    PROTO_FLAVOR = PbModelFlavor.MODEL_FLAVOR_HUGGINGFACE

    HF_TYPE_FILE = "model.hf_type"

    def save_model_to_directory(
        self,
        model_object: ModelObject,
        directory: Path,
    ) -> None:
        model_object.save_pretrained(directory.as_posix())  # type: ignore

        with open(directory / self.HF_TYPE_FILE, "w") as f:
            f.write(type(model_object).__name__)

    def load_model_from_directory(self, directory: Path) -> ModelRuntimeObjects:
        with open(directory / self.HF_TYPE_FILE) as f:
            transformer_type = f.readlines()[0]

            mod = __import__("transformers", fromlist=[transformer_type])
            architecture_class = getattr(mod, transformer_type)

            model = architecture_class.from_pretrained(directory.as_posix())

            return ModelRuntimeObjects(
                model, lambda input_df: self.__predict(model, input_df)
            )

    @staticmethod
    def __predict(model: ModelObject, input_df: pd.DataFrame) -> pd.DataFrame:
        raise Exception("Not implemented")



import hashlib
import inspect
import os
import pickle  # nosec import_pickle
import shutil
import sys
import uuid
from dataclasses import dataclass, field, replace
from pathlib import Path
from typing import Any, Callable, Iterator, List, Optional, Sequence, Tuple, TypeVar

import cloudpickle  # type: ignore
from layerapi.api.entity.run_pb2 import Run as PBRun
from layerapi.api.ids_pb2 import RunId

from layer.config import DEFAULT_FUNC_PATH
from layer.contracts.assertions import Assertion

from .asset import AssetPath, AssetType
from .fabrics import Fabric
from .project_full_name import ProjectFullName


def _language_version() -> Tuple[int, int, int]:
    return sys.version_info.major, sys.version_info.minor, sys.version_info.micro


GetRunsFunction = Callable[[], List[PBRun]]


@dataclass(frozen=True)
class ResourcePath:
    # Local file system path of the resource (file or dir), relative to the project dir.
    # Examples: data/test.csv, users.parquet
    path: str

    def local_relative_paths(self) -> Iterator[str]:
        """
        Map path to the absolute file system paths, checking if paths exist.
        Includes single files and files in each resource directory.

        :return: iterator of absolute file paths.
        """
        file_path = os.path.relpath(self.path)
        if not os.path.exists(file_path):
            raise FileNotFoundError(
                f"resource file or directory: {self.path} in {os.getcwd()}"
            )
        if os.path.isfile(file_path):
            yield file_path
        if os.path.isdir(file_path):
            for root, _, files in os.walk(file_path):
                for f in files:
                    dir_file_path = os.path.join(root, f)
                    yield os.path.relpath(dir_file_path)


FunctionDefinitionType = TypeVar(  # pylint: disable=invalid-name
    "FunctionDefinitionType", bound="FunctionDefinition"
)


class FunctionDefinition:
    def __init__(
        self,
        func: Callable[..., Any],
        project_name: str,
        account_name: str,
        asset_type: AssetType,
        asset_name: str,
        fabric: Fabric,
        asset_dependencies: List[AssetPath],
        pip_dependencies: List[str],
        resource_paths: List[ResourcePath],
        assertions: List[Assertion],
        version_id: Optional[uuid.UUID] = None,
        repository_id: Optional[uuid.UUID] = None,
        description: str = "",
        uri: str = "",
    ) -> None:
        self.func = func
        self.func_name: str = func.__name__

        self.project_name = project_name
        self.account_name = account_name
        self.asset_type = asset_type
        self.asset_name = asset_name
        self.fabric = fabric
        self.asset_dependencies = asset_dependencies
        self.pip_dependencies = pip_dependencies
        self.resource_paths = resource_paths
        self.assertions = assertions

        self.version_id = version_id
        self.repository_id = repository_id
        self.description = description
        self.uri = uri

        self.func_source = inspect.getsource(self.func)

        self.source_code_digest = hashlib.sha256()
        self.source_code_digest.update(self.func_source.encode("utf-8"))

        self._pack()

    def __repr__(self) -> str:
        return f"FunctionDefinition({self.asset_type}, {self.asset_name})"

    @property
    def project_full_name(self) -> ProjectFullName:
        return ProjectFullName(
            account_name=self.account_name,
            project_name=self.project_name,
        )

    @property
    def asset_path(self) -> AssetPath:
        return AssetPath(
            asset_type=self.asset_type,
            asset_name=self.asset_name,
            project_name=self.project_name,
            org_name=self.account_name,
        )

    @property
    def function_home_dir(self) -> Path:
        return DEFAULT_FUNC_PATH / self.asset_path.path()

    def set_version_id(self, version_id: uuid.UUID) -> None:
        self.version_id = version_id

    def set_repository_id(self, repository_id: uuid.UUID) -> None:
        self.repository_id = repository_id

    def drop_dependencies(self) -> "FunctionDefinition":
        self.asset_dependencies = []
        return self

    # DEPRECATED below, will remove once we build the simplified backend
    def get_pickled_function(self) -> bytes:
        return cloudpickle.dumps(self.func, protocol=pickle.DEFAULT_PROTOCOL)

    @property
    def entrypoint(self) -> str:
        return f"{self.asset_name}.pkl"

    @property
    def pickle_path(self) -> Path:
        return self.function_home_dir / self.entrypoint

    @property
    def environment(self) -> str:
        return "requirements.txt"

    @property
    def environment_path(self) -> Path:
        return self.function_home_dir / self.environment

    def get_fabric(self, is_local: bool) -> str:
        if is_local:
            return Fabric.F_LOCAL.value
        else:
            return self.fabric.value

    def _clean_pickle_folder(self) -> None:
        # Remove directory to clean leftovers from previous runs
        function_home_dir = self.function_home_dir
        if function_home_dir.exists():
            shutil.rmtree(function_home_dir)
        os.makedirs(function_home_dir)

    def _pack(self) -> None:
        self._clean_pickle_folder()

        # Dump pickled function to asset_name.pkl
        with open(self.pickle_path, mode="wb") as file:
            cloudpickle.dump(self.func, file, protocol=pickle.DEFAULT_PROTOCOL)

        with open(self.environment_path, "w") as reqs_file:
            reqs_file.write("\n".join(self.pip_dependencies))


@dataclass(frozen=True)
class Run:
    """
    Provides access to project runs stored in Layer.

    You can retrieve an instance of this object with :code:`layer.run()`.

    This class should not be initialized by end-users.

    .. code-block:: python

        # Runs the current project with the given functions
        layer.run([build_dataset, train_model])

    """

    project_full_name: ProjectFullName
    definitions: Sequence[FunctionDefinition] = field(default_factory=list, repr=False)
    files_hash: str = ""
    readme: str = field(repr=False, default="")
    run_id: Optional[RunId] = field(repr=False, default=None)

    @property
    def project_name(self) -> str:
        return self.project_full_name.project_name

    @property
    def account_name(self) -> str:
        return self.project_full_name.account_name

    def with_run_id(self, run_id: RunId) -> "Run":
        return replace(self, run_id=run_id)

    def with_definitions(self, definitions: Sequence[FunctionDefinition]) -> "Run":
        return replace(self, definitions=definitions)



import json
import os
import time
from dataclasses import dataclass, field, replace
from datetime import datetime
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Sequence

import jwt
from yarl import URL

from layer.exceptions.exceptions import (
    ConfigError,
    InvalidConfigurationError,
    MissingConfigurationError,
)
from layer.utils.session import UserSessionId


DEFAULT_LAYER_PATH = Path(os.getenv("LAYER_DEFAULT_PATH", Path.home() / ".layer"))
DEFAULT_PATH = DEFAULT_LAYER_PATH / "config.json"
DEFAULT_LOGS_DIR = DEFAULT_LAYER_PATH / "logs"
DEFAULT_FUNC_PATH = DEFAULT_LAYER_PATH / "functions"
DEFAULT_URL = URL("https://app.layer.ai")
ROOT_LAYER_URL = URL("https://layer.co")


@dataclass(frozen=True)
class ServiceConfig:
    address: str = ""

    @property
    def is_enabled(self) -> bool:
        return bool(self.address)


@dataclass(frozen=True)
class ProjectServiceConfig(ServiceConfig):
    pass


@dataclass(frozen=True)
class DataCatalogConfig(ServiceConfig):
    pass


@dataclass(frozen=True)
class ModelCatalogConfig(ServiceConfig):
    pass


@dataclass(frozen=True)
class ModelTrainingConfig(ServiceConfig):
    pass


@dataclass(frozen=True)
class AccountServiceConfig(ServiceConfig):
    pass


@dataclass(frozen=True)
class FlowManagerServiceConfig(ServiceConfig):
    pass


@dataclass(frozen=True)
class UserLogsServiceConfig(ServiceConfig):
    @property
    def max_receive_message_length(self) -> int:
        return 100 * 1024 * 1024


@dataclass(frozen=True)
class S3Config:
    endpoint_url: Optional[URL] = None

    @classmethod
    def create_default(cls) -> "S3Config":
        return cls()


def _default_logs_file_path() -> Path:
    local_now = datetime.now().strftime("%Y%m%dT%H%M%S")
    logs_file_path = Path.joinpath(
        DEFAULT_LOGS_DIR, f"{local_now}-session-{UserSessionId()}.log"
    )
    return logs_file_path


@dataclass(frozen=True)
class LogsConfig:
    logs_file_path: Path = _default_logs_file_path()


@dataclass(frozen=True)
class ClientConfig:
    data_catalog: DataCatalogConfig
    model_catalog: ModelCatalogConfig
    model_training: ModelTrainingConfig
    account_service: AccountServiceConfig
    flow_manager: FlowManagerServiceConfig
    user_logs: UserLogsServiceConfig
    project_service: ProjectServiceConfig
    grpc_gateway_address: str = ""
    access_token: str = ""
    grpc_do_verify_ssl: bool = True
    logs_file_path: Path = LogsConfig().logs_file_path
    s3: S3Config = S3Config.create_default()

    def with_access_token(self, access_token: str) -> "ClientConfig":
        return replace(self, access_token=access_token)


@dataclass(frozen=True)
class AuthConfig:
    auth_url: URL
    token_url: URL
    logout_url: URL

    client_id: str
    audience: str

    headless_callback_url: URL

    callback_urls: Sequence[URL]
    success_redirect_url: URL
    failure_redirect_url: URL

    @classmethod
    def create_disabled(cls) -> "AuthConfig":
        return AuthConfig(
            auth_url=URL(),
            token_url=URL(),
            logout_url=URL(),
            client_id="",
            audience="",
            headless_callback_url=URL(),
            callback_urls=[],
            success_redirect_url=URL(),
            failure_redirect_url=URL(),
        )

    @property
    def is_enabled(self) -> bool:
        return bool(self.auth_url)

    @property
    def callback_host(self) -> str:
        assert self.callback_urls[0].host
        return self.callback_urls[0].host

    @property
    def callback_ports(self) -> List[int]:
        return [url.port for url in self.callback_urls if url.port]


@dataclass(frozen=True)
class Credentials:
    access_token: str = field(repr=False)
    refresh_token: str = field(repr=False)

    # This is used to assume the tokens expire a bit earlier than their actual
    # expiration time. Therefore we will never send an expired token.
    _expiration_margin: ClassVar[float] = 60.0 * 60.0  # 1 hour

    @classmethod
    def create_empty(cls) -> "Credentials":
        return cls(access_token="", refresh_token="")

    @property
    def _access_token_expiration_time(self) -> float:
        return (
            float(
                jwt.decode(
                    self.access_token,
                    options={"verify_signature": False},
                    algorithms=["HS256"],
                ).get("exp", float("inf"))
            )
            - self._expiration_margin
        )

    @property
    def is_empty(self) -> bool:
        return not bool(self.access_token)

    @property
    def is_access_token_expired(self) -> bool:
        return self.is_empty or time.time() >= self._access_token_expiration_time

    @property
    def is_authenticated_outside_organization(self) -> bool:
        # todo: update this to account_id along with LAY-2716
        return self.is_empty or f"{ROOT_LAYER_URL}/organization_id" not in jwt.decode(
            self.access_token, options={"verify_signature": False}, algorithms=["HS256"]
        )


def get_config(name: str, record: Dict[str, Any]) -> Any:
    if name not in record:
        raise ConfigError(
            "Missing configuration parameter. Make sure you have the latest release and login again."
        )
    return record[name]


def get_config_or_default(name: str, default_val: Any, record: Dict[str, Any]) -> Any:
    if name not in record:
        return default_val
    return record[name]


@dataclass(frozen=True)
class Config:
    url: URL
    client: ClientConfig
    auth: AuthConfig
    credentials: Credentials = Credentials.create_empty()
    is_guest: bool = False

    def with_credentials(self, creds: Credentials) -> "Config":
        return replace(
            self,
            credentials=creds,
            client=replace(self.client, access_token=creds.access_token),
        )


class ConfigRecord:
    @classmethod
    def from_auth(cls, config: AuthConfig) -> Dict[str, Any]:
        if not config.is_enabled:
            return {}
        return {
            "auth_url": str(config.auth_url),
            "token_url": str(config.token_url),
            "logout_url": str(config.logout_url),
            "client_id": config.client_id,
            "audience": config.audience,
            "headless_callback_url": str(config.headless_callback_url),
            "callback_urls": [str(url) for url in config.callback_urls],
            "success_redirect_url": str(config.success_redirect_url),
            "failure_redirect_url": str(config.failure_redirect_url),
        }

    @classmethod
    def to_auth(cls, record: Dict[str, Any]) -> AuthConfig:
        if not record:
            return AuthConfig.create_disabled()
        auth_url = URL(record["auth_url"])
        logout_url = auth_url.with_path("/v2/logout")
        if record.get("logout_url"):
            logout_url = URL(record["logout_url"])
        headless_callback_url = URL(record["headless_callback_url"])
        failure_redirect_url = headless_callback_url
        if record.get("failure_redirect_url"):
            failure_redirect_url = URL(record["failure_redirect_url"])
        return AuthConfig(
            auth_url=URL(record["auth_url"]),
            token_url=URL(record["token_url"]),
            logout_url=logout_url,
            client_id=record["client_id"],
            audience=record["audience"],
            headless_callback_url=headless_callback_url,
            callback_urls=[URL(url) for url in record["callback_urls"]],
            success_redirect_url=URL(record["success_redirect_url"]),
            failure_redirect_url=failure_redirect_url,
        )

    @classmethod
    def from_credentials(cls, creds: Credentials) -> Dict[str, Any]:
        return {
            "access_token": creds.access_token,
            "refresh_token": creds.refresh_token,
        }

    @classmethod
    def to_credentials(cls, record: Dict[str, Any]) -> Credentials:
        if not record:
            return Credentials.create_empty()
        return Credentials(
            access_token=record["access_token"],
            refresh_token=record["refresh_token"],
        )

    @classmethod
    def from_client(cls, config: ClientConfig) -> Dict[str, Any]:
        record: Dict[str, Any] = {
            "grpc_gateway_address": config.grpc_gateway_address,
        }
        if not config.grpc_do_verify_ssl:
            record["grpc_do_verify_ssl"] = config.grpc_do_verify_ssl
        if config.s3.endpoint_url:
            record["s3_endpoint_url"] = str(config.s3.endpoint_url)
        return record

    @classmethod
    def to_client(cls, record: Dict[str, Any], access_token: str) -> ClientConfig:
        grpc_gateway_address = get_config("grpc_gateway_address", record)
        grpc_do_verify_ssl = record.get("grpc_do_verify_ssl", True)
        if "s3_endpoint_url" in record:
            s3_config = S3Config(endpoint_url=URL(record["s3_endpoint_url"]))
        else:
            s3_config = S3Config.create_default()
        return ClientConfig(
            data_catalog=DataCatalogConfig(address=grpc_gateway_address),
            model_catalog=ModelCatalogConfig(address=grpc_gateway_address),
            model_training=ModelTrainingConfig(address=grpc_gateway_address),
            account_service=AccountServiceConfig(address=grpc_gateway_address),
            flow_manager=FlowManagerServiceConfig(address=grpc_gateway_address),
            user_logs=UserLogsServiceConfig(address=grpc_gateway_address),
            project_service=ProjectServiceConfig(address=grpc_gateway_address),
            grpc_gateway_address=grpc_gateway_address,
            access_token=access_token,
            grpc_do_verify_ssl=grpc_do_verify_ssl,
            s3=s3_config,
        )

    @classmethod
    def from_config(cls, config: Config) -> Dict[str, Any]:
        assert config.credentials
        return {
            "is_guest": config.is_guest,
            "url": str(config.url),
            "auth": cls.from_auth(config.auth),
            "credentials": cls.from_credentials(config.credentials),
            "client": cls.from_client(config.client),
        }

    @classmethod
    def to_config(cls, record: Dict[str, Any]) -> Config:
        is_guest = bool(get_config_or_default("is_guest", False, record))
        url = URL(get_config("url", record))
        creds = cls.to_credentials(get_config("credentials", record))
        auth = cls.to_auth(get_config("auth", record))
        client = cls.to_client(
            record=get_config("client", record), access_token=creds.access_token
        )
        return Config(
            is_guest=is_guest,
            url=url,
            auth=auth,
            credentials=creds,
            client=client,
        )


class ConfigStore:
    def __init__(self, path: Path) -> None:
        self._path = path

    def save(self, config: Config) -> None:
        record = ConfigRecord.from_config(config)
        self._path.parent.mkdir(parents=True, exist_ok=True)
        with open(self._path, "w") as f:
            json.dump(record, f)
        self._path.chmod(0o600)

    def load(self) -> Config:
        try:
            with open(self._path, "r") as f:
                return ConfigRecord.to_config(json.load(f))
        except IOError:
            raise MissingConfigurationError(self._path)
        except Exception:
            raise InvalidConfigurationError(self._path)

    def delete(self) -> None:
        try:
            self._path.unlink()  # same as unlink(self, missing_ok=True), but works on Python < 3.8
        except FileNotFoundError:
            pass



from typing import Any, Dict, List, Optional, Tuple, Union

import wrapt  # type: ignore

from layer.contracts.assets import AssetPath, AssetType, BaseAsset
from layer.contracts.datasets import Dataset
from layer.contracts.models import Model
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.runs import FunctionDefinition
from layer.projects.utils import get_current_project_full_name
from layer.settings import LayerSettings


# See https://wrapt.readthedocs.io/en/latest/wrappers.html#custom-function-wrappers for more.
class LayerFunctionWrapper(wrapt.FunctionWrapper):
    def __init__(
        self,
        wrapped: Any,
        wrapper: Any,
        enabled: Any,
    ) -> None:
        super().__init__(wrapped, wrapper, enabled)
        if not hasattr(wrapped, "layer"):
            wrapped.layer = LayerSettings()
        self.layer: LayerSettings = wrapped.layer

    # wrapt doesn't implement this method and based on this https://github.com/GrahamDumpleton/wrapt/issues/102#issuecomment-899937490
    # we give it a shot and it seems to be working
    def __reduce_ex__(self, protocol: Any) -> Any:
        return type(self), (self.__wrapped__, self._self_wrapper, self._self_enabled)

    def __copy__(self) -> None:
        pass

    def __deepcopy__(self, memo: Dict[int, object]) -> None:
        pass

    def __reduce__(self) -> Union[str, Tuple[Any, ...]]:
        pass


class LayerAssetFunctionWrapper(LayerFunctionWrapper):
    def __init__(
        self,
        wrapped: Any,
        wrapper: Any,
        enabled: Any,
        asset_type: AssetType,
        name: str,
        dependencies: Optional[List[Union[str, Dataset, Model]]],
    ) -> None:
        super().__init__(wrapped, wrapper, enabled)
        self.layer.set_asset_type(asset_type)
        self.layer.set_asset_name(name)

        paths: List[AssetPath] = []
        if dependencies is not None:
            for dependency in dependencies:
                if isinstance(dependency, str):
                    paths.append(AssetPath.parse(dependency))
                elif isinstance(dependency, BaseAsset):
                    paths.append(dependency._path)
                else:
                    raise ValueError(
                        "Dependencies can only be a string, Dataset or Model."
                    )
        self.layer.set_dependencies(paths)

    def get_definition(self) -> FunctionDefinition:
        self.layer.validate()
        current_project_full_name = get_current_project_full_name()

        # get asset dependencies
        asset_dependencies: List[AssetPath] = []
        for d in self.layer.get_dependencies():
            full_path_dep = d
            if d.is_relative():
                if d.project_name is not None:
                    full_path_dep = d.with_project_full_name(
                        ProjectFullName(
                            account_name=current_project_full_name.account_name,
                            project_name=d.project_name,
                        )
                    )
                else:
                    full_path_dep = d.with_project_full_name(current_project_full_name)
            asset_dependencies.append(full_path_dep)

        # get pip dependencies
        pip_dependencies: List[str] = []
        if self.layer.get_pip_requirements_file():
            with open(self.layer.get_pip_requirements_file(), "r") as file:
                pip_dependencies = file.read().strip().split("\n")
        else:
            pip_dependencies = self.layer.get_pip_packages()

        return FunctionDefinition(
            func=self.__wrapped__,
            project_name=current_project_full_name.project_name,
            account_name=current_project_full_name.account_name,
            asset_type=self.layer.get_asset_type(),
            asset_name=self.layer.get_asset_name(),
            fabric=self.layer.get_fabric(),
            asset_dependencies=asset_dependencies,
            pip_dependencies=pip_dependencies,
            resource_paths=self.layer.get_resource_paths(),
            assertions=self.layer.get_assertions(),
        )



import json
import ssl
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import grpc

from .interceptors import (
    GRPCErrorClientInterceptor,
    LogRpcCallsInterceptor,
    RequestIdInterceptor,
)


def create_grpc_channel(
    address: str,
    access_token: str,
    *,
    do_verify_ssl: bool = True,
    logs_file_path: Path,
    options: Optional[List[Tuple[str, Union[str, int]]]] = None,
) -> Any:
    # https://grpc.github.io/grpc/cpp/md_doc_keepalive.html
    # https://github.com/grpc/proposal/blob/master/A8-client-side-keepalive.md
    if options is None:
        options = []
    options = options.copy()
    ssl_config = create_grpc_ssl_config(address, do_verify_ssl=do_verify_ssl)
    if ssl_config.hostname_override:
        options.append(("grpc.ssl_target_name_override", ssl_config.hostname_override))
    json_config = json.dumps(
        {
            "methodConfig": [
                {
                    "name": [{}],
                    "retryPolicy": {
                        "maxAttempts": 5,
                        "initialBackoff": "0.1s",
                        "maxBackoff": "5s",
                        "backoffMultiplier": 2,
                        "retryableStatusCodes": ["UNAVAILABLE"],
                    },
                }
            ]
        }
    )
    options.append(("grpc.enable_retries", 1))
    options.append(("grpc.service_config", json_config))
    options.append(("grpc.keepalive_time_ms", 60000))
    options.append(("grpc.keepalive_timeout_ms", 5000))
    options.append(("grpc.keepalive_permit_without_calls", 1))
    options.append(("grpc.http2.max_pings_without_data", 0))
    credentials = grpc.ssl_channel_credentials(ssl_config.cadata)

    client_interceptors = [
        RequestIdInterceptor(),
        GRPCErrorClientInterceptor(),
        LogRpcCallsInterceptor(logs_file_path),
    ]

    return grpc.intercept_channel(
        grpc.secure_channel(
            address,
            grpc.composite_channel_credentials(
                credentials,
                grpc.access_token_call_credentials(access_token),
            ),
            options,
        ),
        *client_interceptors,
    )


@dataclass(frozen=True)
class GRPCSSLConfig:
    cadata: Optional[bytes] = None
    hostname_override: Optional[str] = None


def _load_default_ssl_certs() -> bytes:
    ssl_context = ssl.create_default_context()
    der_certs = ssl_context.get_ca_certs(binary_form=True)  # certs are in der format
    pem_certs = [ssl.DER_cert_to_PEM_cert(der_cert) for der_cert in der_certs]  # type: ignore
    return "\n".join(pem_certs).encode()


def create_grpc_ssl_config(
    address: str, *, do_verify_ssl: bool = True, do_force_cadata_load: bool = False
) -> GRPCSSLConfig:
    from cryptography import x509
    from cryptography.x509.oid import ExtensionOID

    cadata: Optional[bytes] = None

    if do_verify_ssl:
        if do_force_cadata_load:
            cadata = _load_default_ssl_certs()
        return GRPCSSLConfig(cadata=cadata)

    host, port = address.split(":", 1)
    cadata = ssl.get_server_certificate((host, int(port))).encode()
    cert = x509.load_pem_x509_certificate(cadata)
    hostname_override = cert.extensions.get_extension_for_oid(
        ExtensionOID.SUBJECT_ALTERNATIVE_NAME
    ).value.get_values_for_type(x509.DNSName)[0]
    return GRPCSSLConfig(cadata=cadata, hostname_override=hostname_override)



from typing import Any, Callable, Dict, List, Optional

import wrapt  # type: ignore

from layer.decorators.layer_wrapper import LayerFunctionWrapper


def pip_requirements(
    file: Optional[str] = None, packages: Optional[List[str]] = None
) -> Callable[..., Any]:
    """
    Allows specifying requirements to install in Layer backend when running decorated function.

    Only one of file or packages can be defined. These settings will override requirements set via ``layer.init()``.

    The decorator should only be used alongside Layer ``dataset`` or ``model`` decorators. It will only have effect on the remote execution of your decorated functions. If you require specific packages and you want to run the decorated function locally, you will need to install those requirements locally yourself.

    :param file: Path to file listing requirements to install.
    :param packages: List of packages (and optionally their version) to install in Layer backend when running decorated function.
    :return: Function object.

    .. code-block:: python

        from layer.decorators import dataset, pip_requirements

        layer.init("your-project-name", pip_packages=["tensorflow"])

        # You can specify libraries to install via pip_requirements. Doing so will override
        # requirements specified via layer.init above when running `create_product_dataset`
        @pip_requirements(packages=["pandas==1.4.1"])
        @dataset("product-data")
        def create_product_dataset():
            data = [[1, "product1", 15], [2, "product2", 20], [3, "product3", 10]]
            dataframe = pd.DataFrame(data, columns=["Id", "Product", "Price"])
            return dataframe
    """
    if file and packages:
        raise ValueError("either file or packages should be provided, not both")

    @wrapt.decorator(proxy=_pip_requirements_wrapper(file, packages))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _pip_requirements_wrapper(
    file: Optional[str], packages: Optional[List[str]]
) -> Any:
    class PipRequirementsFunctionWrapper(LayerFunctionWrapper):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any) -> None:
            super().__init__(wrapped, wrapper, enabled)
            self.layer.set_pip_requirements_file(file)
            self.layer.set_pip_packages(packages)

    return PipRequirementsFunctionWrapper



from typing import TYPE_CHECKING, Dict, List, Optional

from layer.types import ModelObject

from .base import ModelFlavor  # noqa
from .catboost import CatBoostModelFlavor  # noqa
from .custom import CustomModelFlavor  # noqa
from .huggingface import HuggingFaceModelFlavor  # noqa
from .keras import KerasModelFlavor  # noqa
from .lightgbm import LightGBMModelFlavor  # noqa
from .pytorch import PyTorchModelFlavor  # noqa
from .sklearn import ScikitLearnModelFlavor  # noqa
from .tensorflow import TensorFlowModelFlavor  # noqa
from .xgboost import XGBoostModelFlavor  # noqa


if TYPE_CHECKING:
    from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PbModelFlavor


# Order matters for matching
PYTHON_FLAVORS: List[ModelFlavor] = [
    # HF Flavor should come before Pytorch and Tensorflow since it uses Pytorch and Tensorflow as back bone
    HuggingFaceModelFlavor(),
    KerasModelFlavor(),
    PyTorchModelFlavor(),
    TensorFlowModelFlavor(),
    # XGB Flavor should come before sklearn since it uses it as back bone
    XGBoostModelFlavor(),
    ScikitLearnModelFlavor(),
    LightGBMModelFlavor(),
    CatBoostModelFlavor(),
    CustomModelFlavor(),
]

# mypy proto enum typing issue https://github.com/protocolbuffers/protobuf/issues/8175
PYTHON_CLASS_NAME_TO_PROTO_FLAVORS: Dict[str, "PbModelFlavor.ValueType"] = {
    flavor.__class__.__name__: flavor.PROTO_FLAVOR for flavor in PYTHON_FLAVORS
}

PROTO_TO_PYTHON_OBJECT_FLAVORS: Dict["PbModelFlavor.ValueType", ModelFlavor] = {
    flavor.PROTO_FLAVOR: flavor for flavor in PYTHON_FLAVORS
}


def get_flavor_for_model(model_object: ModelObject) -> Optional[ModelFlavor]:
    matching_flavor: Optional[ModelFlavor] = None
    for flavor in PYTHON_FLAVORS:
        if flavor.can_interpret_object(model_object):
            matching_flavor = flavor
            break
    return matching_flavor


def get_flavor_for_proto(
    proto_flavor: "PbModelFlavor.ValueType",
) -> Optional[ModelFlavor]:
    return PROTO_TO_PYTHON_OBJECT_FLAVORS.get(proto_flavor)



from dataclasses import dataclass
from typing import List, Optional, Union

from .context import Context
from .contracts.fabrics import Fabric
from .contracts.project_full_name import ProjectFullName


@dataclass
class GlobalContext:
    project_full_name: Optional[ProjectFullName]
    fabric: Optional[Fabric]
    active_context: Optional[Context]
    pip_requirements_file: Optional[str]
    pip_packages: Optional[List[str]]
    # We show a message to the user if their installed layer version is outdated.
    # We want to avoid showing this in case we already checked the version or we have already shown this message.
    has_shown_update_message: bool
    # Similar to above, but for supported Python version.
    has_shown_python_version_message: bool


# We store project name, fabric, active context and requirements
# at the process level for use in subsequent calls in the same Python process.
_GLOBAL_CONTEXT = GlobalContext(
    project_full_name=None,
    fabric=None,
    active_context=None,
    pip_requirements_file=None,
    pip_packages=None,
    has_shown_update_message=False,
    has_shown_python_version_message=False,
)


def reset_to(project_full_name: Optional[Union[str, ProjectFullName]]) -> None:
    project_full_name = _project_full_name_from(project_full_name)
    if current_project_full_name() != project_full_name:
        global _GLOBAL_CONTEXT
        _GLOBAL_CONTEXT = GlobalContext(
            project_full_name=project_full_name,
            fabric=None,
            active_context=None,
            pip_requirements_file=None,
            pip_packages=None,
            has_shown_update_message=False,
            has_shown_python_version_message=False,
        )


def set_current_project_full_name(name: Optional[Union[str, ProjectFullName]]) -> None:
    _GLOBAL_CONTEXT.project_full_name = _project_full_name_from(name)


def _project_full_name_from(
    name: Optional[Union[str, ProjectFullName]]
) -> Optional[ProjectFullName]:
    if not name:
        return None
    if isinstance(name, ProjectFullName):
        return name

    parts = name.split("/")
    if len(parts) != 2:
        raise ValueError(
            "Project full name badly formatted. Use 'account-name/project-name' "
            "or ProjectFullName(account_name='account-name', project_name='project-name')"
        )
    account_name, project_name = parts
    return ProjectFullName(
        account_name=account_name,
        project_name=project_name,
    )


def current_project_full_name() -> Optional[ProjectFullName]:
    return _GLOBAL_CONTEXT.project_full_name


def current_project_name() -> Optional[str]:
    return (
        _GLOBAL_CONTEXT.project_full_name.project_name
        if _GLOBAL_CONTEXT.project_full_name
        else None
    )


def current_account_name() -> Optional[str]:
    return (
        _GLOBAL_CONTEXT.project_full_name.account_name
        if _GLOBAL_CONTEXT.project_full_name
        else None
    )


def set_active_context(context: Context) -> None:
    _GLOBAL_CONTEXT.active_context = context


def reset_active_context() -> None:
    _GLOBAL_CONTEXT.active_context = None


def get_active_context() -> Optional[Context]:
    """
    Returns the active context object set from the active computation. Used in local mode to identify which
    context to log resources to.

    @return:  active context object
    """
    return _GLOBAL_CONTEXT.active_context


def set_default_fabric(fabric: Fabric) -> None:
    _GLOBAL_CONTEXT.fabric = fabric


def default_fabric() -> Optional[Fabric]:
    return _GLOBAL_CONTEXT.fabric


def set_pip_requirements_file(pip_requirements_file: str) -> None:
    _GLOBAL_CONTEXT.pip_requirements_file = pip_requirements_file


def get_pip_requirements_file() -> Optional[str]:
    return _GLOBAL_CONTEXT.pip_requirements_file


def set_pip_packages(packages: List[str]) -> None:
    _GLOBAL_CONTEXT.pip_packages = packages


def get_pip_packages() -> Optional[List[str]]:
    return _GLOBAL_CONTEXT.pip_packages


def set_has_shown_update_message(shown: bool) -> None:
    _GLOBAL_CONTEXT.has_shown_update_message = shown


def has_shown_update_message() -> bool:
    return _GLOBAL_CONTEXT.has_shown_update_message


def set_has_shown_python_version_message(shown: bool) -> None:
    _GLOBAL_CONTEXT.has_shown_python_version_message = shown


def has_shown_python_version_message() -> bool:
    return _GLOBAL_CONTEXT.has_shown_python_version_message



from layerapi.api.service.account.user_api_pb2 import GetGuestAuthTokenRequest
from layerapi.api.service.account.user_api_pb2_grpc import UserAPIStub

from layer.config import LogsConfig
from layer.utils.grpc import create_grpc_channel


def get_guest_auth_token(url: str) -> str:
    with create_grpc_channel(
        url, "", logs_file_path=LogsConfig().logs_file_path
    ) as channel:
        svc = UserAPIStub(channel=channel)
        auth_token = svc.GetGuestAuthToken(GetGuestAuthTokenRequest())
        return auth_token.token



import sys
import tempfile
import uuid
from contextlib import contextmanager
from logging import Logger
from pathlib import Path
from typing import Any, Callable, Generator, Iterator, List, Optional, Sequence, Tuple

import pandas
import pyarrow
from layerapi.api.entity.dataset_build_pb2 import DatasetBuild as PBDatasetBuild
from layerapi.api.entity.dataset_list_options_pb2 import (
    DatasetListOptions,
    DatasetSortField,
)
from layerapi.api.entity.dataset_pb2 import Dataset as PBDataset
from layerapi.api.entity.dataset_version_pb2 import DatasetVersion as PBDatasetVersion
from layerapi.api.ids_pb2 import DatasetBuildId, DatasetId, DatasetVersionId, ProjectId
from layerapi.api.service.datacatalog.data_catalog_api_pb2 import (
    CompleteBuildRequest,
    CompleteBuildResponse,
    GetBuildRequest,
    GetDatasetRequest,
    GetDatasetsRequest,
    GetLatestBuildRequest,
    GetPythonDatasetAccessCredentialsRequest,
    GetPythonDatasetAccessCredentialsResponse,
    GetResourcePathsRequest,
    GetVersionRequest,
    InitiateBuildRequest,
    InitiateBuildResponse,
    RegisterDatasetRequest,
    UpdateResourcePathsIndexRequest,
)
from layerapi.api.service.datacatalog.data_catalog_api_pb2_grpc import (
    DataCatalogAPIStub,
)
from layerapi.api.service.dataset.dataset_api_pb2 import (
    Command,
    DatasetQuery,
    DatasetSnapshot,
)
from layerapi.api.value.language_version_pb2 import LanguageVersion
from layerapi.api.value.python_dataset_pb2 import PythonDataset as PBPythonDataset
from layerapi.api.value.python_source_pb2 import PythonSource
from layerapi.api.value.s3_path_pb2 import S3Path
from layerapi.api.value.storage_location_pb2 import StorageLocation
from layerapi.api.value.ticket_pb2 import DatasetPathTicket, DataTicket
from pyarrow import flight

from layer.config import ClientConfig
from layer.contracts.assets import AssetPath, AssetType
from layer.contracts.datasets import (
    Dataset,
    DatasetBuild,
    DatasetBuildStatus,
    SortField,
)
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.runs import FunctionDefinition
from layer.exceptions.exceptions import LayerClientException
from layer.pandas_extensions import _infer_custom_types
from layer.utils.file_utils import tar_directory
from layer.utils.grpc import create_grpc_channel, generate_client_error_from_grpc_error
from layer.utils.s3 import S3Util

from .dataset_service import DatasetClient, DatasetClientError


class DataCatalogClient:
    _service: DataCatalogAPIStub

    def __init__(
        self,
        config: ClientConfig,
        logger: Logger,
        *,
        service_factory: Callable[..., DataCatalogAPIStub] = DataCatalogAPIStub,
        flight_client: flight.FlightClient = None,
        dataset_client: Optional["DatasetClient"] = None,
    ):
        self._config = config.data_catalog
        self._logger = logger
        self._service_factory = service_factory
        self._access_token = config.access_token
        self._call_metadata = [("authorization", f"Bearer {config.access_token}")]
        self._flight_client = flight_client
        self._do_verify_ssl = config.grpc_do_verify_ssl
        self._logs_file_path = config.logs_file_path
        self._s3_endpoint_url = config.s3.endpoint_url
        self._dataset_client = (
            DatasetClient(
                address_and_port=config.grpc_gateway_address,
                access_token=config.access_token,
            )
            if dataset_client is None
            else dataset_client
        )

    @contextmanager
    def init(self) -> Iterator["DataCatalogClient"]:
        with create_grpc_channel(
            self._config.address,
            self._access_token,
            do_verify_ssl=self._do_verify_ssl,
            logs_file_path=self._logs_file_path,
        ) as channel:
            self._service = self._service_factory(channel)
            yield self

    def _get_python_dataset_access_credentials(
        self, dataset_path: AssetPath
    ) -> GetPythonDatasetAccessCredentialsResponse:
        return self._service.GetPythonDatasetAccessCredentials(
            GetPythonDatasetAccessCredentialsRequest(dataset_path=dataset_path.path()),
        )

    def fetch_dataset(
        self, asset_path: AssetPath, no_cache: bool = False
    ) -> "pandas.DataFrame":
        data_ticket = DataTicket(
            dataset_path_ticket=DatasetPathTicket(path=asset_path.path()),
        )
        command = Command(dataset_query=DatasetQuery(ticket=data_ticket))
        all_partition_data = []
        try:
            for partition_metadata in self._dataset_client.get_partitions_metadata(
                command
            ):
                partition = self._dataset_client.fetch_partition(
                    partition_metadata, no_cache=no_cache
                )
                all_partition_data.append(partition.to_pandas())
        except DatasetClientError as e:
            raise LayerClientException(str(e))

        if len(all_partition_data) > 0:
            df = pandas.concat(all_partition_data, ignore_index=True)
        else:
            df = pandas.DataFrame()

        return df

    def _get_dataset_writer(self, name: str, build_id: uuid.UUID, schema: Any) -> Any:
        dataset_snapshot = DatasetSnapshot(build_id=DatasetBuildId(value=str(build_id)))
        return self._dataset_client.get_dataset_writer(
            Command(dataset_snapshot=dataset_snapshot), schema
        )

    def store_dataset(
        self,
        name: str,
        data: "pandas.DataFrame",
        build_id: uuid.UUID,
        progress_callback: Optional[Callable[[int], None]] = None,
    ) -> None:
        """
        Store a dataset.

        :param name: the name of the dataset
            If empty, the data is written to internal storage
        :param data: dataset data
        :param build_id: dataset build id
        :param progress_callback: progress callback
        """

        # Creates a Record batch from the pandas dataframe
        batch = pyarrow.RecordBatch.from_pandas(
            _infer_custom_types(data), preserve_index=False
        )
        try:
            writer, _ = self._get_dataset_writer(name, build_id, batch.schema)
            try:
                total_rows = 0  # total rows written
                for chunk in _get_batch_chunks(batch):
                    writer.write_batch(chunk)
                    total_rows += chunk.num_rows
                    if progress_callback:
                        progress_callback(batch.num_rows - total_rows)
            finally:
                writer.close()
        except Exception as err:
            raise generate_client_error_from_grpc_error(
                err, "internal dataset store error"
            )

    def initiate_build(
        self,
        dataset: FunctionDefinition,
        project_id: uuid.UUID,
        is_local: bool,
    ) -> InitiateBuildResponse:
        self._logger.debug(
            "Initiating build for the dataset %r",
            dataset.asset_name,
        )

        resp = self._service.InitiateBuild(
            InitiateBuildRequest(
                dataset_name=dataset.asset_name,
                format="python",
                build_entity_type=PBDatasetBuild.BUILD_ENTITY_TYPE_DATASET,
                project_id=ProjectId(value=str(project_id)),
                fabric=dataset.get_fabric(is_local),
            )
        )

        return resp

    def complete_build(
        self,
        dataset_build_id: DatasetBuildId,
        dataset: FunctionDefinition,
        error: Optional[Exception] = None,
    ) -> CompleteBuildResponse:
        self._logger.debug(
            "Completing build for the dataset %r",
            dataset.asset_name,
        )

        if error:
            max_error_length = 99_999
            placeholder = "..."
            raw_text = str.format("Dataset build failed with {}", error)
            status = DatasetBuildStatus.FAILED
            success = None
            failure = CompleteBuildRequest.BuildFailed(
                info=(raw_text[: (max_error_length - len(placeholder))] + placeholder)
                if len(raw_text) > max_error_length
                else raw_text
            )
        else:
            status = DatasetBuildStatus.COMPLETED
            success = CompleteBuildRequest.BuildSuccess(
                location=StorageLocation(uri=dataset.uri), schema="{}"
            )
            failure = None

        resp = self._service.CompleteBuild(
            CompleteBuildRequest(
                id=dataset_build_id,
                status=status,
                success=success,
                failure=failure,
            )
        )

        return resp

    def add_dataset(
        self,
        project_id: uuid.UUID,
        dataset_definition: FunctionDefinition,
        is_local: bool,
    ) -> str:
        self._logger.debug(
            "Adding or updating a dataset with name %r",
            dataset_definition.asset_name,
        )
        resp = self._service.RegisterDataset(
            RegisterDatasetRequest(
                name=dataset_definition.asset_name,
                description=dataset_definition.description,
                python_dataset=self._get_pb_python_dataset(
                    dataset_definition, is_local
                ),
                project_id=ProjectId(value=str(project_id)),
            ),
        )
        return resp.dataset_id.value

    def _get_pb_python_dataset(
        self,
        dataset: FunctionDefinition,
        is_local: bool,
    ) -> PBPythonDataset:
        s3_path = self._upload_dataset_source(dataset)
        language_version = _language_version()
        return PBPythonDataset(
            s3_path=s3_path,
            python_source=PythonSource(
                content=dataset.func_source,
                entrypoint=dataset.entrypoint,
                environment=dataset.environment,
                language_version=LanguageVersion(
                    major=language_version[0],
                    minor=language_version[1],
                    micro=language_version[2],
                ),
            ),
            fabric=dataset.get_fabric(is_local),
        )

    def _upload_dataset_source(self, dataset: FunctionDefinition) -> S3Path:
        response = self._get_python_dataset_access_credentials(dataset.asset_path)
        archive_name = f"{dataset.asset_name}.tgz"

        with tempfile.TemporaryDirectory() as tmp_dir:
            archive_path = f"{tmp_dir}/{archive_name}"
            tar_directory(archive_path, dataset.function_home_dir)
            S3Util.upload_dir(
                Path(tmp_dir),
                response.credentials,
                response.s3_path,
                endpoint_url=self._s3_endpoint_url,
            )
        return S3Path(
            bucket=response.s3_path.bucket,
            key=f"{response.s3_path.key}{archive_name}",
        )

    def get_resource_paths(
        self, project_full_name: ProjectFullName, function_name: str, path: str = ""
    ) -> List[str]:
        request = GetResourcePathsRequest(
            project_full_name=project_full_name.path,
            function_name=function_name,
            path=path,
        )
        response = self._service.GetResourcePaths(request)
        return response.paths

    def update_resource_paths_index(
        self, project_full_name: ProjectFullName, function_name: str, paths: List[str]
    ) -> None:
        request = UpdateResourcePathsIndexRequest(
            project_full_name=project_full_name.path,
            function_name=function_name,
            paths=paths,
        )
        self._service.UpdateResourcePathsIndex(request)

    def get_dataset_by_id(self, id_: uuid.UUID) -> Dataset:
        dataset = self._get_dataset_by_id(str(id_))
        build = self._get_build_by_id(dataset.default_build_id.value)
        version = self._get_version_by_id(build.dataset_version_id.value)
        return self._create_dataset(dataset, version, build)

    def get_dataset_by_name(
        self, project_id: uuid.UUID, name: str, version_name: str = ""
    ) -> Dataset:
        build = self._get_build_by_name(project_id, name, version_name)
        version = self._get_version_by_id(build.dataset_version_id.value)
        dataset = self._get_dataset_by_id(version.dataset_id.value)
        return self._create_dataset(dataset, version, build)

    def get_dataset_by_build_id(self, id_: uuid.UUID) -> Dataset:
        build = self._get_build_by_id(str(id_))
        version = self._get_version_by_id(build.dataset_version_id.value)
        dataset = self._get_dataset_by_id(version.dataset_id.value)
        return self._create_dataset(dataset, version, build)

    def list_datasets(
        self,
        sort_fields: Sequence[SortField] = (),
        query_build: bool = True,
    ) -> Iterator[Dataset]:
        for dataset in self._list_datasets(sort_fields):
            if query_build:
                build = self._get_build_by_id(dataset.default_build_id.value)
                version = self._get_version_by_id(build.dataset_version_id.value)
            else:
                build = PBDatasetBuild()
                version = PBDatasetVersion()
            yield self._create_dataset(dataset, version, build)

    def _create_dataset(
        self,
        dataset: PBDataset,
        version: PBDatasetVersion,
        build: PBDatasetBuild,
    ) -> Dataset:

        asset_path = AssetPath(
            asset_type=AssetType.DATASET,
            asset_name=dataset.name,
            asset_version=version.name,
            project_name=dataset.project_name,
        )
        return Dataset(
            id=uuid.UUID(dataset.id.value),
            asset_path=asset_path,
            description=dataset.description,
            schema=version.schema,
            version=version.name,
            uri=build.location.uri,
            metadata=dict(build.location.metadata.value),
            build=DatasetBuild(
                id=build.id.value,
                status=DatasetBuildStatus(build.build_info.status),
                info=build.build_info.info,
                index=str(build.index),
            ),
        )

    def _get_dataset_by_id(self, id_: str) -> PBDataset:
        return self._service.GetDataset(
            GetDatasetRequest(dataset_id=DatasetId(value=id_)),
        ).dataset

    def _get_version_by_id(self, id_: str) -> PBDatasetVersion:
        return self._service.GetVersion(
            GetVersionRequest(version_id=DatasetVersionId(value=id_)),
        ).version

    def _get_build_by_id(self, id_: str) -> PBDatasetBuild:
        return self._service.GetBuild(
            GetBuildRequest(build_id=DatasetBuildId(value=id_)),
        ).build

    def _get_build_by_name(
        self, project_id: uuid.UUID, name: str, version: str = ""
    ) -> PBDatasetBuild:
        return self._service.GetLatestBuild(
            GetLatestBuildRequest(
                dataset_name=name,
                dataset_version=version,
                project_id=ProjectId(value=str(project_id)),
            ),
        ).build

    def _list_datasets(self, sort_fields: Sequence[SortField] = ()) -> List[PBDataset]:
        return list(
            self._service.GetDatasets(
                GetDatasetsRequest(
                    dataset_list_options=DatasetListOptions(
                        sorting=[
                            DatasetSortField(
                                name=sort_field.name, descending=sort_field.descending
                            )
                            for sort_field in sort_fields
                        ]
                    )
                ),
            ).datasets
        )


def _get_batch_chunks(
    batch: pyarrow.RecordBatch, max_chunk_size_bytes: int = 4_000_000
) -> Generator[pyarrow.RecordBatch, None, None]:
    """
    Slice the batch into chunks, based on average row size,
    but not exceeding the maximum chunk size.
    """

    bytes_per_row = batch.nbytes / batch.num_rows  # average row size
    if bytes_per_row > max_chunk_size_bytes:
        raise ValueError(
            f"the average size of a single row of {int(bytes_per_row)} byte(s) exceeds max chunk size of {max_chunk_size_bytes} byte(s)"
        )
    max_num_rows = int(
        max_chunk_size_bytes / bytes_per_row
    )  # how many rows fit into the chunk
    start = 0
    while start < batch.num_rows:
        i = max_num_rows
        # in case chunk size is exceeded due to large individual rows, slim down until it fits
        while i > 0:
            chunk = batch.slice(start, i)
            if chunk.nbytes <= max_chunk_size_bytes:
                yield chunk
                break
            i -= 1
        if i == 0:
            raise ValueError(
                f"single row in the batch at index {start} exceeds max chunk size of {max_chunk_size_bytes} byte(s)"
            )
        start += i


def _language_version() -> Tuple[int, int, int]:
    return sys.version_info.major, sys.version_info.minor, sys.version_info.micro



import pickle  # nosec
from abc import abstractmethod
from pathlib import Path
from typing import Any

import pandas
from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PbModelFlavor

from layer.types import ModelObject

from .base import ModelFlavor, ModelRuntimeObjects


class CustomModel:
    """
    A generic model that evaluates inputs and produces outputs.
    """

    def __init__(self) -> None:
        """
        Initializes the custom  model
        """

    @abstractmethod
    def predict(self, model_input: pandas.DataFrame) -> pandas.DataFrame:
        """
        Evaluates an input for this model and produces an output.

        :param model_input: A pandas.DataFrame as input for the model to evaluate.
        :return: Model output as a pandas.DataFrame
        """


class CustomModelFlavor(ModelFlavor):
    """A model flavor implementation which enables custom model implementation for Layer users."""

    MODULE_KEYWORD = "layer"
    MODEL_PICKLE_FILE = "custom_model.pkl"
    MODEL_SOURCE_FILE = "custom_model.py"
    MODEL_CONFIG_FILE = "custom_model.config"
    PROTO_FLAVOR = PbModelFlavor.MODEL_FLAVOR_CUSTOM

    def save_model_to_directory(
        self,
        model_object: Any,
        directory: Path,
    ) -> None:
        """Saves a model to a directory."""
        directory.mkdir(parents=True, exist_ok=True)

        # Store class config
        with open(directory / self.MODEL_CONFIG_FILE, mode="w") as file:
            custom_model_class = type(model_object)
            class_config = (
                custom_model_class.__module__ + "." + custom_model_class.__name__
            )
            file.write(class_config)

        # Store class definition
        import inspect

        custom_class_model = type(model_object)
        try:
            class_model_source_code = inspect.getsource(custom_class_model)
        except OSError:
            raise Exception(
                "Can't retrieve the source code of your custom model. Implement your model in a separate python file"
            )
        with open(directory / self.MODEL_SOURCE_FILE, mode="w") as file:
            # Plain pickle doesn't pickle the loaded modules at the root level. Since we expect users to inherit
            # `layer.CustomModel`, we add it at the top of the model source code
            class_model_source_code = (
                "import layer\n"
                "from layer import CustomModel\n"
                "\n" + class_model_source_code
            )
            file.write(class_model_source_code)

        # Store model itself
        with open(directory / self.MODEL_PICKLE_FILE, mode="wb") as file:
            try:
                pickle.dump(
                    model_object, file, protocol=pickle.DEFAULT_PROTOCOL
                )  # nosec
            except AttributeError:
                raise Exception(
                    "You should define your custom model in a separate python file"
                )

    def load_model_from_directory(self, directory: Path) -> ModelRuntimeObjects:
        """Loads a model from a directory.

        Returns:
            Model object and prediction function.
        """

        # Load model config
        with open(directory / self.MODEL_CONFIG_FILE, mode="r") as file:
            model_config = file.readline()
            model_name = model_config.split(".")[-1]
            model_module_name = ".".join(model_config.split(".")[:-1])

        # Load and register custom class definition
        import importlib.util
        import sys

        try:
            spec = importlib.util.spec_from_file_location(
                model_name, directory / self.MODEL_SOURCE_FILE
            )
            module = importlib.util.module_from_spec(spec)  # type: ignore
            spec.loader.exec_module(module)  # type: ignore
            sys.modules[model_module_name] = module
        except Exception:
            raise Exception(
                "Failed to load model. Please clear your cache with `layer.clear_cache` and try again!"
            )

        # Load model itself
        with open(directory / self.MODEL_PICKLE_FILE, mode="rb") as file:
            model = pickle.load(file)  # nosec
            return ModelRuntimeObjects(
                model, lambda input_df: self.__predict(model, input_df)
            )

    @staticmethod
    def __predict(model: ModelObject, input_df: pandas.DataFrame) -> pandas.DataFrame:
        return model.predict(input_df)  # type: ignore



import inspect
from abc import ABCMeta, abstractmethod, abstractproperty
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional

import pandas as pd

from layer.types import ModelObject


if TYPE_CHECKING:
    from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PBModelFlavor


@dataclass(frozen=True)
class ModelRuntimeObjects:
    model_object: ModelObject
    prediction_function: Optional[Callable[[pd.DataFrame], pd.DataFrame]] = None


class ModelFlavor(metaclass=ABCMeta):
    """Represents a machine learning model flavor for a specific framework.

    Implementations provide methods for checking for membership, saving and loading
    of machine learning models within their flavor.
    """

    @abstractproperty
    def MODULE_KEYWORD(self) -> str:  # pylint: disable=invalid-name
        """Defines a keyword as part of an object's module name that matches this flavor.

        Returns:
            keyword: a str
        """

    @abstractproperty
    def PROTO_FLAVOR(  # pylint: disable=invalid-name
        self,
    ) -> "PBModelFlavor.ValueType":
        """Defines the proto flavor that this Model Flavor uses.

        Returns:
            The proto flavor
        """

    def can_interpret_object(self, model_object: ModelObject) -> bool:
        """Checks whether supplied model object has flavor of this class.

        Args:
            model_object: A machine learning model which could be originated from any framework.

        Returns:
            bool: true if this ModelFlavor can interpret the given model instance.
        """
        for hierarchy_class in inspect.getmro(type(model_object)):
            parent_module = inspect.getmodule(hierarchy_class)
            if (
                parent_module is not None
                and self.MODULE_KEYWORD in parent_module.__name__
            ):
                return True

        return False

    @abstractmethod
    def save_model_to_directory(
        self,
        model_object: ModelObject,
        directory: Path,
    ) -> None:
        """Defines the method that this Model Flavor uses to save a model to a directory."""

    @abstractmethod
    def load_model_from_directory(self, directory: Path) -> ModelRuntimeObjects:
        """Defines the method that this Model Flavor uses to load a model from a directory.

        Returns:
             Model object and prediction function.
        """



from pathlib import Path

import pandas as pd
from layerapi.api.value.model_flavor_pb2 import ModelFlavor as PbModelFlavor

from layer.types import ModelObject

from .base import ModelFlavor, ModelRuntimeObjects


LAYER_CACHED_TF_MODELS = {}


class TensorFlowModelFlavor(ModelFlavor):
    """An ML Model flavor implementation which handles persistence of TensorFlow Models."""

    MODULE_KEYWORD = "tensorflow"
    PROTO_FLAVOR = PbModelFlavor.MODEL_FLAVOR_TENSORFLOW

    def save_model_to_directory(
        self, model_object: ModelObject, directory: Path
    ) -> None:
        import mlflow.tensorflow
        import tensorflow  # type: ignore

        tmp_save_path = directory / "tensorflowmodel"
        tensorflow.saved_model.save(model_object, tmp_save_path)

        mlflow.tensorflow.save_model(
            tf_saved_model_dir=tmp_save_path,
            tf_meta_graph_tags=None,
            tf_signature_def_key="serving_default",
            path=directory.as_posix(),
        )

    def load_model_from_directory(self, directory: Path) -> ModelRuntimeObjects:
        import mlflow.tensorflow

        model = mlflow.tensorflow.load_model(directory.as_uri())
        return ModelRuntimeObjects(
            model, lambda input_df: self.__predict(directory, input_df)
        )

    @staticmethod
    def __predict(directory: Path, input_df: pd.DataFrame) -> pd.DataFrame:
        cache_key = str(directory)
        if cache_key not in LAYER_CACHED_TF_MODELS:
            import mlflow.pyfunc

            model = mlflow.pyfunc.load_model(directory.as_uri())
            LAYER_CACHED_TF_MODELS[cache_key] = model
        model = LAYER_CACHED_TF_MODELS[cache_key]
        return model.predict(input_df)



from .base import ModelFlavor  # noqa
from .catboost import CatBoostModelFlavor  # noqa
from .custom import CustomModelFlavor  # noqa
from .huggingface import HuggingFaceModelFlavor  # noqa
from .keras import KerasModelFlavor  # noqa
from .lightgbm import LightGBMModelFlavor  # noqa
from .pytorch import PyTorchModelFlavor  # noqa
from .sklearn import ScikitLearnModelFlavor  # noqa
from .tensorflow import TensorFlowModelFlavor  # noqa
from .xgboost import XGBoostModelFlavor  # noqa



import logging
from typing import Any, Callable, Dict, List, Optional, Union
from uuid import UUID

import wrapt  # type: ignore

from layer import Dataset, Model
from layer.clients.layer import LayerClient
from layer.config import ConfigManager
from layer.config.config import Config
from layer.contracts.assets import AssetType
from layer.contracts.runs import FunctionDefinition
from layer.decorators.layer_wrapper import LayerAssetFunctionWrapper
from layer.projects.utils import (
    get_current_project_full_name,
    verify_project_exists_and_retrieve_project_id,
)
from layer.tracker.progress_tracker import RunProgressTracker
from layer.training.runtime.model_train_failure_reporter import (
    ModelTrainFailureReporter,
)
from layer.utils.async_utils import asyncio_run_in_thread


logger = logging.getLogger(__name__)


def model(
    name: str, dependencies: Optional[List[Union[str, Dataset, Model]]] = None
) -> Callable[..., Any]:
    """
    Decorator function used to wrap another function that trains a model. The function that decorator has been applied to needs to return a ML model object from a supported framework.

    The decorator ensures that:
    - If decorated function is executed locally, then the model trained is stored in the Layer backend.
    - If decorated function is passed in to `layer.run(your_function)`, then Layer runs the function remotely and stores the model in Layer backend.

    :param name: Name with which the model are stored in Layer's backend.
    :param dependencies: List of datasets or models that will be built by Layer backend prior to building the current function. Layer understand what entities this function depends on so it optimizes the build process accordingly.
    :return: Function object being decorated.

    .. code-block:: python

        from layer
        from layer.decorators import model

        # Define a new function for model generation

        @model("model_1")
        def create_model_1():
            from sklearn import datasets
            from sklearn.svm import SVC

            iris = datasets.load_iris()
            clf = SVC()
            return clf.fit(iris.data, iris.target)

        # below will train model using Layer's backend
        layer.run([create_model_1])

    If you use a dataset that is outside of your project, then you must explicitly call it out as a dependency.
    These dependencies are displayed in the **External assets** section in your Layer project.

    The dataset used in the model example below is from the a public Layer project.

    .. code-block:: python

        import pandas as pd
        from layer
        from layer.decorators import dataset, model

        @model(name='survival_model', dependencies=[layer.Dataset('layer/titanic/datasets/features')])
        @assert_true(test_survival_probability)
        def train():
            df = layer.get_dataset("layer/titanic/datasets/features").to_pandas()
            X = df.drop(["Survived"], axis=1)
            y = df["Survived"]
            X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=42)
            random_forest = RandomForestClassifier(n_estimators=100)
            random_forest.fit(X_train, y_train)
            y_pred = random_forest.predict(X_test)
            return random_forest
    """

    @wrapt.decorator(proxy=_model_wrapper(name, dependencies))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _model_wrapper(
    name: str,
    dependencies: Optional[List[Union[str, Dataset, Model]]] = None,
) -> Any:
    class FunctionWrapper(LayerAssetFunctionWrapper):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any) -> None:
            super().__init__(
                wrapped,
                wrapper,
                enabled,
                AssetType.MODEL,
                name,
                dependencies,
            )

        # This is not serialized with cloudpickle, so it will only be run locally.
        # See https://layerco.slack.com/archives/C02R5B3R3GU/p1646144705414089 for detail.
        def __call__(self, *args: Any, **kwargs: Any) -> Any:
            current_project_full_name_ = get_current_project_full_name()
            config: Config = asyncio_run_in_thread(ConfigManager().refresh())
            with LayerClient(config.client, logger).init() as client:
                progress_tracker = RunProgressTracker(
                    url=config.url,
                    project_name=current_project_full_name_.project_name,
                    account_name=current_project_full_name_.account_name,
                )

                with progress_tracker.track() as tracker:
                    tracker.add_asset(AssetType.MODEL, self.layer.get_asset_name())
                    model_definition = self.get_definition()
                    return self._train_model_locally_and_store_remotely(
                        model_definition, tracker, client
                    )

        @staticmethod
        def _train_model_locally_and_store_remotely(
            model_definition: FunctionDefinition,
            tracker: RunProgressTracker,
            client: LayerClient,
        ) -> Any:
            from layer.training.runtime.model_trainer import (
                LocalTrainContext,
                ModelTrainer,
            )

            assert model_definition.project_name is not None
            verify_project_exists_and_retrieve_project_id(
                client, model_definition.project_full_name
            )

            model_version = client.model_catalog.create_model_version(
                model_definition.project_full_name,
                model_definition,
                True,
            ).model_version
            train_id = client.model_catalog.create_model_train_from_version_id(
                model_version.id
            )
            train = client.model_catalog.get_model_train(train_id)

            context = LocalTrainContext(  # noqa: F841
                logger=logger,
                model_name=model_definition.asset_name,
                model_version=model_version.name,
                train_id=UUID(train_id.value),
                source_folder=model_definition.function_home_dir,
                source_entrypoint=model_definition.entrypoint,
                train_index=str(train.index),
            )
            failure_reporter = ModelTrainFailureReporter(
                client.model_catalog,
                logger,
                context.train_id,
                context.source_folder,
            )
            trainer = ModelTrainer(
                client=client,
                train_context=context,
                logger=logger,
                failure_reporter=failure_reporter,
                tracker=tracker,
            )
            result = trainer.train()

            return result

    return FunctionWrapper



import enum


@enum.unique
class Fabric(enum.Enum):
    F_LOCAL = "f-local"
    F_XXSMALL = "f-xxsmall"
    F_XSMALL = "f-xsmall"
    F_SMALL = "f-small"
    F_MEDIUM = "f-medium"
    F_GPU_SMALL = "f-gpu-small"

    @classmethod
    def has_member_key(cls, key: str) -> bool:
        try:
            cls.__new__(cls, key)
            return True
        except ValueError:
            return False

    @classmethod
    def default(cls) -> "Fabric":
        return Fabric.F_SMALL

    def is_gpu(self) -> bool:
        return "gpu" in self.value



import re
import uuid
from abc import ABCMeta, abstractproperty
from dataclasses import dataclass, replace
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Sequence, TypeVar, Union

from yarl import URL

from layer.cache.cache import Cache
from layer.contracts.project_full_name import ProjectFullName
from layer.exceptions.exceptions import LayerClientException


_ASSET_PATH_PATTERN = re.compile(
    r"^(([a-zA-Z0-9_-]+)\/)?(([a-zA-Z0-9_-]+)\/)?(datasets|models)\/([a-zA-Z0-9_-]+)(:([a-z0-9_]*)(\.([0-9]*))?)?(#([a-zA-Z0-9_-]+))?$"
)

BaseAssetType = TypeVar(  # pylint: disable=invalid-name
    "BaseAssetType", bound="BaseAsset"
)


class AssetType(Enum):
    DATASET = "datasets"
    MODEL = "models"


@dataclass(frozen=True)
class AssetPath:
    asset_name: str
    asset_type: AssetType
    # TODO rename to 'account_name'
    org_name: Optional[str] = None
    project_name: Optional[str] = None
    asset_version: Optional[str] = None
    asset_build: Optional[int] = None
    asset_selector: Optional[str] = None

    @classmethod
    def parse(
        cls,
        composite_name: str,
        expected_asset_type: Optional[AssetType] = None,
    ) -> "AssetPath":
        if len(composite_name.split("/")) < 2:
            if not expected_asset_type:
                raise ValueError("Please specify full path or specify asset type")
            composite_name = f"{expected_asset_type.value}/{composite_name}"

        result = _ASSET_PATH_PATTERN.search(composite_name)
        if not result:
            raise ValueError("Asset path does not match expected pattern")
        groups = result.groups()
        optional_project = groups[3] if groups[3] else groups[1]
        optional_org = groups[1] if groups[3] else None
        maybe_asset_type = groups[4]
        if maybe_asset_type:
            asset_type = AssetType(maybe_asset_type)
        elif expected_asset_type:
            asset_type = expected_asset_type
        else:
            raise ValueError(
                "expected asset type either in the composite name or as argument"
            )
        if asset_type and expected_asset_type and asset_type != expected_asset_type:
            raise ValueError(
                f"expected asset type {expected_asset_type} but found {asset_type}"
            )

        name = groups[5]
        if not name:
            raise ValueError("Asset name missing")
        optional_version = groups[7]
        optional_build = groups[9]
        optional_selector = groups[11]

        return cls(
            asset_name=name,
            asset_type=asset_type,
            asset_version=optional_version,
            asset_build=int(optional_build) if optional_build else None,
            asset_selector=optional_selector,
            project_name=optional_project,
            org_name=optional_org,
        )

    def is_relative(self) -> bool:
        return self.org_name is None or self.project_name is None

    def has_project(self) -> bool:
        return self.project_name is not None and self.project_name != ""

    def path(self) -> str:
        parts = [
            self.org_name,
            self.project_name,
            self.asset_type.value,
            self.asset_name,
        ]
        p = "/".join([part for part in parts if part is not None])
        if self.asset_version is not None:
            p = f"{p}:{self.asset_version}"
            if self.asset_build is not None:
                p = f"{p}.{self.asset_build}"

        if self.asset_selector is not None:
            p = f"{p}#{self.asset_selector}"

        return p

    def with_project_full_name(self, project_full_name: ProjectFullName) -> "AssetPath":
        return replace(
            self,
            project_name=project_full_name.project_name,
            org_name=project_full_name.account_name,
        )

    def url(self, base_url: URL) -> URL:
        if self.org_name is None:
            raise LayerClientException("Account name is required to get URL")
        if self.project_name is None:
            raise LayerClientException("Project name is required to get URL")
        return base_url / self.path()

    def must_org_name(self) -> str:
        if self.org_name is None:
            raise LayerClientException("Account name is required")
        return self.org_name

    def must_project_name(self) -> str:
        if self.project_name is None:
            raise LayerClientException("Project name is required")
        return self.project_name


class BaseAsset(metaclass=ABCMeta):
    def __init__(
        self,
        path: Union[str, AssetPath],
        id: Optional[uuid.UUID] = None,
        dependencies: Optional[Sequence["BaseAsset"]] = None,
    ):
        if dependencies is None:
            dependencies = []
        self._path = (
            AssetPath.parse(path, self.asset_type) if isinstance(path, str) else path
        )
        self._id = id
        self._dependencies = dependencies

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BaseAsset):
            return False

        return (
            self._path == other._path
            and self._id == other._id
            and self._dependencies == other._dependencies
        )

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    @abstractproperty
    def asset_type(self) -> AssetType:
        ...

    @property
    def name(self) -> str:
        return self._path.asset_name

    @property
    def path(self) -> str:
        return self._path.path()

    @property
    def id(self) -> uuid.UUID:
        assert self._id is not None
        return self._id

    @property
    def dependencies(self) -> Sequence["BaseAsset"]:
        return self._dependencies

    @property
    def project_name(self) -> Optional[str]:
        return self._path.project_name

    def with_id(self: BaseAssetType, id: uuid.UUID) -> BaseAssetType:
        self._id = id
        return self

    def with_project_full_name(self, project_full_name: ProjectFullName) -> "BaseAsset":
        new_path = self._path.with_project_full_name(project_full_name)
        return self.__class__(
            path=new_path,
            id=self.id,
            dependencies=self.dependencies,
        )

    def with_dependencies(
        self: BaseAssetType, dependencies: Sequence["BaseAsset"]
    ) -> BaseAssetType:
        self._dependencies = dependencies
        return self

    def drop_dependencies(self: BaseAssetType) -> BaseAssetType:
        return self.with_dependencies(())

    def get_cache_dir(self, cache_dir: Optional[Path] = None) -> Optional[Path]:
        cache = Cache(cache_dir=cache_dir).initialise()
        return cache.get_path_entry(str(self.id))

    def is_cached(self, cache_dir: Optional[Path] = None) -> bool:
        return self.get_cache_dir(cache_dir) is not None



from typing import Any

import pandas as pd


def check_and_convert_to_df(result: Any) -> pd.DataFrame:
    allowed_types = [pd.DataFrame, list]
    if not any([isinstance(result, x_type) for x_type in allowed_types]):
        raise Exception(
            f"Unsupported return type. Expected one of {allowed_types}, got: {type(result)}"
        )
    if isinstance(result, pd.DataFrame):
        if len(result.columns) == 0:
            raise Exception("pandas.DataFrame object has 0 columns")
        return result
    if isinstance(result, list):
        dict_elements = all([isinstance(list_element, dict) for list_element in result])
        if not dict_elements:
            raise Exception("Expected a list of dict elements")
        return pd.DataFrame.from_records(result)  # type: ignore
    raise Exception(
        f"Unsupported return type. Expected one of {allowed_types}, got: {type(result)}"
    )



from contextlib import ExitStack, contextmanager
from logging import Logger
from typing import Iterator

from layer.config import ClientConfig

from .account_service import AccountServiceClient
from .data_catalog import DataCatalogClient
from .flow_manager import FlowManagerClient
from .logged_data_service import LoggedDataClient
from .model_catalog import ModelCatalogClient
from .model_training_service import ModelTrainingClient
from .project_service import ProjectServiceClient
from .user_logs_service import UserLogsClient


class LayerClient:
    def __init__(self, config: ClientConfig, logger: Logger):
        self._config = config
        self._data_catalog = DataCatalogClient(config, logger)
        self._model_catalog = ModelCatalogClient(config, logger)
        self._model_training = ModelTrainingClient(config, logger)
        self._account = AccountServiceClient(config, logger)
        self._flow_manager = FlowManagerClient(config, logger)
        self._user_logs = UserLogsClient(config, logger)
        self._project_service_client = ProjectServiceClient(config, logger)
        self._logged_data_client = LoggedDataClient(config, logger)

    @contextmanager
    def init(self) -> Iterator["LayerClient"]:
        with ExitStack() as exit_stack:
            # TODO(emin): remove unused config objects
            if self._config.data_catalog.is_enabled:
                exit_stack.enter_context(self._data_catalog.init())
            if self._config.model_catalog.is_enabled:
                exit_stack.enter_context(self._model_catalog.init())
            if self._config.model_training.is_enabled:
                exit_stack.enter_context(self._model_training.init())
            if self._config.account_service.is_enabled:
                exit_stack.enter_context(self._account.init())
            if self._config.flow_manager.is_enabled:
                exit_stack.enter_context(self._flow_manager.init())
            if self._config.user_logs.is_enabled:
                exit_stack.enter_context(self._user_logs.init())
            if self._config.project_service.is_enabled:
                exit_stack.enter_context(self.project_service_client.init())
            exit_stack.enter_context(self._logged_data_client.init())
            yield self

    @property
    def data_catalog(self) -> DataCatalogClient:
        return self._data_catalog

    @property
    def model_catalog(self) -> ModelCatalogClient:
        return self._model_catalog

    @property
    def model_training(self) -> ModelTrainingClient:
        return self._model_training

    @property
    def account(self) -> AccountServiceClient:
        return self._account

    @property
    def flow_manager(self) -> FlowManagerClient:
        return self._flow_manager

    @property
    def user_logs(self) -> UserLogsClient:
        return self._user_logs

    @property
    def project_service_client(self) -> ProjectServiceClient:
        return self._project_service_client

    @property
    def logged_data_service_client(self) -> LoggedDataClient:
        return self._logged_data_client



import abc
import asyncio
import base64
import errno
import hashlib
import secrets
import webbrowser
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, AsyncIterator, ClassVar, Sequence
from urllib.parse import urlencode

from yarl import URL

from layer.config import AuthConfig, Credentials
from layer.exceptions.exceptions import AuthException


if TYPE_CHECKING:
    from aiohttp import ClientSession
    from aiohttp.web import Application, Request, Response


EMPTY_URL = URL()


def urlsafe_unpadded_b64encode(payload: bytes) -> str:
    return base64.urlsafe_b64encode(payload).decode().rstrip("=")


@asynccontextmanager
async def create_app_server_once(
    app: "Application", *, host: str = "127.0.0.1", port: int = 8080
) -> AsyncIterator[URL]:
    from aiohttp.web import AppRunner, TCPSite

    runner = AppRunner(app, access_log=None)
    try:
        await runner.setup()
        site = TCPSite(runner, host, port, shutdown_timeout=0.0)
        await site.start()
        yield URL(site.name)
    finally:
        await runner.shutdown()
        await runner.cleanup()


@asynccontextmanager
async def create_app_server(
    app: "Application", *, host: str = "127.0.0.1", ports: Sequence[int] = (8080,)
) -> AsyncIterator[URL]:
    for port in ports:
        try:
            async with create_app_server_once(app, host=host, port=port) as url:
                yield url
            return
        except OSError as err:
            if err.errno != errno.EADDRINUSE:
                raise
    raise RuntimeError("No free ports.")


@dataclass(frozen=True)
class CodeChallenge:
    verifier: str
    value: str
    method: ClassVar[str] = "S256"

    @classmethod
    def create(cls) -> "CodeChallenge":
        verifier = urlsafe_unpadded_b64encode(secrets.token_bytes(32))
        digest = hashlib.sha256(verifier.encode()).digest()
        challenge = urlsafe_unpadded_b64encode(digest)
        return cls(
            verifier=verifier,
            value=challenge,
        )


@dataclass(frozen=True)
class Code:
    challenge: CodeChallenge
    value: str
    callback_url: URL


class CodeCallbackHandler:
    def __init__(
        self,
        future: "asyncio.Future[str]",
        config: AuthConfig,
    ) -> None:
        self._future = future
        self._config = config

    async def handle(self, request: "Request") -> "Response":
        from aiohttp.web import HTTPBadRequest, HTTPFound, Response

        if "error" in request.query:
            await self._handle_error(request)

        code = request.query.get("code")

        if not code:
            self._future.cancel()
            raise HTTPBadRequest(text="The 'code' query parameter is missing.")

        if not self._future.cancelled():
            self._future.set_result(code)

        if self._config.success_redirect_url:
            raise HTTPFound(self._config.success_redirect_url)
        return Response(text="OK")

    async def _handle_error(self, request: "Request") -> None:
        from aiohttp.web import HTTPFound

        error = request.query["error"]
        description = request.query.get("error_description", "")

        if not self._future.cancelled():
            self._future.set_exception(AuthException(description))

        raise HTTPFound(self._generate_logout_url(error, description))

    def _generate_logout_url(self, error: str, description: str) -> URL:
        query = {"client_id": self._config.client_id}
        if self._config.failure_redirect_url:
            query["returnTo"] = str(
                self._config.failure_redirect_url.with_query(
                    error=error, error_description=description
                )
            )
        return self._config.logout_url.with_query(**query)


def create_auth_code_app(
    future: "asyncio.Future[str]",
    config: AuthConfig,
) -> "Application":
    from aiohttp.web import Application

    app = Application()
    handler = CodeCallbackHandler(future, config)
    app.router.add_get("/", handler.handle)
    return app


class CodeClient(abc.ABC):
    def __init__(self, *, config: AuthConfig) -> None:
        self._config = config

    def _generate_auth_url(
        self, code_challenge: CodeChallenge, callback_url: URL
    ) -> URL:
        return self._config.auth_url.with_query(
            response_type="code",
            code_challenge=code_challenge.value,
            code_challenge_method=code_challenge.method,
            client_id=self._config.client_id,
            redirect_uri=str(callback_url),
            scope="offline_access",
            audience=self._config.audience,
        )

    @abc.abstractmethod
    async def request(self) -> Code:
        pass


class WebBrowserCodeClient(CodeClient):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._timeout_s = kwargs.pop("timeout_s", 5 * 60.0)
        self._code_callback = kwargs.pop("callback", webbrowser.open_new)
        super().__init__(*args, **kwargs)

    async def request(self) -> Code:
        code_challenge = CodeChallenge.create()
        loop = asyncio.get_event_loop()
        future: asyncio.Future[str] = asyncio.Future()
        app = create_auth_code_app(future, self._config)
        try:
            async with create_app_server(
                app, host=self._config.callback_host, ports=self._config.callback_ports
            ) as callback_url:
                auth_url = self._generate_auth_url(code_challenge, callback_url)
                await loop.run_in_executor(None, self._code_callback, str(auth_url))
                try:
                    await asyncio.wait_for(future, self._timeout_s)
                except (asyncio.TimeoutError, asyncio.CancelledError) as err:
                    raise AuthException("failed to get an authorization code") from err
        except (OSError, RuntimeError) as err:
            raise AuthException("failed to start a code callback server") from err
        return Code(
            challenge=code_challenge, value=future.result(), callback_url=callback_url
        )


class HeadlessCodeClient(CodeClient):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._code_callback = kwargs.pop("callback", input)
        super().__init__(*args, **kwargs)

    async def request(self) -> Code:
        code_challenge = CodeChallenge.create()
        loop = asyncio.get_event_loop()
        callback_url = self._config.headless_callback_url
        auth_url = self._generate_auth_url(code_challenge, callback_url)
        print(
            "Please open the following link in your web browser. "
            "Once logged in, copy the code and paste it here."
        )
        print(auth_url)
        try:
            value = await loop.run_in_executor(None, self._code_callback, "Code: ")
            if not value:
                raise ValueError("Empty code")
        except (EOFError, KeyboardInterrupt, ValueError) as exc:
            raise AuthException("No code was provided") from exc
        return Code(challenge=code_challenge, value=value, callback_url=callback_url)


class CredentialsClient:
    def __init__(self, *, client: "ClientSession", url: URL, client_id: str) -> None:
        self._client = client

        self._url = url
        self._client_id = client_id

    async def request(self, code: Code) -> Credentials:
        from aiohttp import ClientResponseError

        payload = {
            "grant_type": "authorization_code",
            "code_verifier": code.challenge.verifier,
            "code": code.value,
            "client_id": self._client_id,
            "redirect_uri": str(code.callback_url),
        }
        async with self._client.post(
            self._url,
            headers={
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded",
            },
            data=urlencode(payload),
        ) as resp:
            try:
                resp.raise_for_status()
            except ClientResponseError as exc:
                raise AuthException("failed to get an access token.") from exc
            resp_payload = await resp.json()
            return Credentials(
                access_token=resp_payload["access_token"],
                refresh_token=resp_payload["refresh_token"],
            )

    async def refresh(self, creds: Credentials) -> Credentials:
        from aiohttp import ClientResponseError

        payload = {
            "grant_type": "refresh_token",
            "refresh_token": creds.refresh_token,
            "client_id": self._client_id,
        }
        async with self._client.post(
            self._url,
            headers={
                "accept": "application/json",
                "content-type": "application/x-www-form-urlencoded",
            },
            data=urlencode(payload),
        ) as resp:
            try:
                resp.raise_for_status()
            except ClientResponseError as exc:
                raise AuthException("failed to get an access token.") from exc
            resp_payload = await resp.json()
            return Credentials(
                access_token=resp_payload["access_token"],
                refresh_token=creds.refresh_token,
            )



import logging
import sys
import traceback
from logging import Logger
from typing import Callable

import grpc
import wrapt  # type: ignore

from .exceptions import RuntimeMemoryException


default_logger = logging.getLogger(__name__)


def exception_handler(stage: str, callback=None, logger: Logger = default_logger) -> Callable:  # type: ignore
    """
    Common utility to handle exceptions and report them to their respective backends via callbacks.

    :param stage: at which stage exception has been occurred.
    :param callback: a callback which would be called in case of exception to report it to a backend.
    If none provided and decorated function has callback keyword argument it would be used.
    :param logger: logger to log exceptions in case of structured logging configured for a module.
    Otherwise the default logger would be used.
    """

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):  # type: ignore
        callback_override = callback if callback else kwargs.pop("callback", None)
        try:
            return wrapped(*args, **kwargs)
        except grpc.RpcError as rpc_error_call:
            failure_exc = _transform_grpc_exception(rpc_error_call)
            _log_exception_from_execution(logger)
        except Exception as e:
            _log_exception_from_execution(logger)
            failure_exc = e
            if isinstance(e, MemoryError):
                failure_exc = _transform_memory_error_exception(e)
        except SystemExit as e:
            _log_exception_from_execution(logger)
            failure_exc = _transform_system_exit_error(e)
        try:
            if callback_override:
                callback_override(stage, failure_exc)
        except Exception as e:
            reason = getattr(e, "message", repr(e))
            logger.error(
                f"Failure during reporting the failure: ({failure_exc}) : {reason}",
            )
            _log_exception_from_execution(logger)
        finally:
            sys.exit(1)

    return wrapper


def _transform_system_exit_error(e: SystemExit) -> Exception:
    # To populate stacktrace
    try:
        raise Exception(f"System exit with code:{e.code}")
    except Exception as ee:
        return ee


def _transform_memory_error_exception(e: MemoryError) -> Exception:
    # To populate stacktrace
    try:
        raise RuntimeMemoryException(str(e))
    except Exception as ee:
        return ee


def _transform_grpc_exception(exc: grpc.RpcError) -> Exception:
    """
    Done to populate sys.exc_info() with an exception containing the stacktrace
    of the input param exception, so that traceback.print_exc(), which we use
    to print debug logs, would print the complete exception stacktrace as intended.
    :param exc: The exception, whose error message we want to extract + retain its stacktrace
    """
    failure_reason = exc.details()
    try:
        raise Exception(failure_reason) from exc
    except Exception as e:
        return e


def _log_exception_from_execution(logger: Logger) -> None:
    logger.warning("Caught exception stacktrace:", exc_info=True)
    traceback.print_exc(file=sys.stdout)



import abc
import uuid
from contextlib import contextmanager
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from layerapi.api.ids_pb2 import RunId
from rich.progress import Task
from rich.text import Text
from yarl import URL

from layer.contracts.assertions import Assertion
from layer.contracts.asset import AssetPath, AssetType
from layer.contracts.tracker import (
    AssetTracker,
    AssetTrackerStatus,
    DatasetTransferState,
    ResourceTransferState,
)
from layer.exceptions.exceptions import ProjectBaseException, ProjectRunnerError

from .output import get_progress_ui


class RunProgressTracker(abc.ABC):
    def __init__(
        self,
        url: URL,
        account_name: str,
        project_name: str,
        assets: Optional[List[Tuple[AssetType, str]]] = None,
    ):
        self._url = url
        self._account_name = account_name
        self._project_name = project_name
        self._assets = assets
        self._progress = get_progress_ui()
        self._tasks: Dict[Tuple[AssetType, str], Task] = {}

    @staticmethod
    def __google_colab_ipykernel_fix() -> None:
        """
        Fixes https://linear.app/layer/issue/LAY-3286/replace-rich-as-a-dependency-for-the-ui-of-the-sdk
        It works by clearing the logger handlers as some of them appear to be interacting with the same resources
        rich interacts with, leading to hitting a bug in old ipykernel-s(<5.2) (https://github.com/ipython/ipykernel/pull/463)
        """
        import logging.config

        logging.config.dictConfig(
            {
                "version": 1,
                "disable_existing_loggers": False,
            }
        )

    @contextmanager
    def track(self) -> Iterator["RunProgressTracker"]:
        """
        Initializes tracking. Meant to be used with a `with` construct.
        """
        self.__google_colab_ipykernel_fix()
        with self._progress:
            self._init_tasks()
            yield self

    def _get_url(self, asset_type: AssetType, name: str) -> URL:
        return AssetPath(
            asset_name=name,
            asset_type=asset_type,
            org_name=self._account_name,
            project_name=self._project_name,
        ).url(self._url)

    def _init_tasks(self) -> None:
        if self._assets:
            for asset in self._assets:
                self._get_or_create_task(
                    asset_type=asset[0],
                    asset_name=asset[1],
                )

    def _get_or_create_task(self, asset_type: AssetType, asset_name: str) -> Task:
        task_key = (asset_type, asset_name)
        if task_key not in self._tasks:
            task_id = self._progress.add_task(
                status=AssetTrackerStatus.PENDING,
                start=False,
                asset=AssetTracker(
                    type=asset_type, name=asset_name, status=AssetTrackerStatus.PENDING
                ),
                description="pending",
            )
            task = self._progress._tasks[task_id]  # pylint: disable=protected-access
            self._tasks[task_key] = task
        return self._tasks[task_key]

    def add_asset(self, asset_type: AssetType, asset_name: str) -> None:
        self._get_or_create_task(asset_type, asset_name)

    def _get_asset(self, asset_type: AssetType, asset_name: str) -> AssetTracker:
        task = self._get_or_create_task(asset_type, asset_name)
        asset = task.fields["asset"]
        return asset

    def _update_asset(
        self,
        asset_type: AssetType,
        asset_name: str,
        *,
        status: Optional[AssetTrackerStatus] = None,
        cur_step: int = 0,
        total_steps: int = 0,
        url: Optional[URL] = None,
        version: Optional[str] = None,
        build_idx: Optional[str] = None,
        error_reason: str = "",
        description: Optional[str] = None,
        model_transfer_state: Optional[ResourceTransferState] = None,
        dataset_transfer_state: Optional[DatasetTransferState] = None,
        resource_transfer_state: Optional[ResourceTransferState] = None,
        loading_cache_asset: Optional[str] = None,
        asset_download_transfer_state: Optional[
            Union[ResourceTransferState, DatasetTransferState]
        ] = None,
    ) -> None:
        task = self._get_or_create_task(asset_type, asset_name)
        asset: AssetTracker = task.fields["asset"]
        if status:
            asset.status = status
        if url:
            asset.base_url = url
        if version:
            asset.version = version
        if build_idx:
            asset.build_idx = build_idx
        if error_reason:
            asset.error_reason = error_reason
        if dataset_transfer_state:
            asset.dataset_transfer_state = dataset_transfer_state
        if model_transfer_state:
            asset.model_transfer_state = model_transfer_state
        if resource_transfer_state:
            asset.resource_transfer_state = resource_transfer_state
        if loading_cache_asset:
            asset.loading_cache_asset = loading_cache_asset
        if asset_download_transfer_state:
            asset.asset_download_transfer_state = asset_download_transfer_state

        progress_args: Dict[str, Any] = {
            "description": description if description is not None else status,
        }

        if status == AssetTrackerStatus.SAVING:
            if asset_type == AssetType.DATASET:
                # Even if we go through all steps, we still want to keep track of time elapsed until the status is DONE, so we add +1 to total_steps.
                progress_args["completed"] = cur_step
                progress_args["total"] = total_steps + 1
                progress_args["description"] = f"saved {cur_step}/{total_steps} rows"
        elif status in (
            AssetTrackerStatus.ASSET_DOWNLOADING,
            AssetTrackerStatus.ASSET_FROM_CACHE,
        ):
            if not task.started:
                task.start_time = task.stop_time = None
                self._progress.start_task(task.id)
                assert task.total is not None
                progress_args["completed"] = task.total * 0.1
        elif status in (
            AssetTrackerStatus.BUILDING,
            AssetTrackerStatus.TRAINING,
        ):
            if not task.started:
                self._progress.start_task(task.id)
        elif status == AssetTrackerStatus.DONE:
            progress_args["completed"] = task.total
            self._progress.stop_task(task.id)
        elif status == AssetTrackerStatus.ERROR:
            self._progress.stop_task(task.id)
            progress_args["completed"] = 0
        self._progress.update(task.id, **progress_args)

    def mark_error_messages(
        self, exc: Union[ProjectBaseException, ProjectRunnerError]
    ) -> None:
        self._progress.stop()  # Stops the progress of non-failed tasks
        self._progress.print(
            Text.from_markup(":x:").append(
                Text(
                    " Project execution error.",
                    style="bold red",
                )
            )
        )
        self._progress.print(
            Text.from_markup(":mag_right:").append(
                Text.from_markup(
                    f" {exc.error_msg_rich if isinstance(exc, ProjectBaseException) else str(exc)}"
                )
            )
        )
        if isinstance(exc, ProjectBaseException):
            self._progress.print(
                Text.from_markup(":bulb:").append(
                    Text.from_markup(
                        f" {exc.suggestion_rich}",
                    )
                )
            )

    def mark_start_running(self, run_id: RunId) -> None:
        pass

    def mark_dataset_saved(self, name: str, *, id_: uuid.UUID) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            url=self._get_url(AssetType.DATASET, name),
            status=AssetTrackerStatus.PENDING,
        )

    def mark_dataset_building(
        self, name: str, version: Optional[str] = None, build_idx: Optional[str] = None
    ) -> None:
        # For some reason, this function is called even after the dataset is successfully built and
        # it causes a flicker in the progress bar. This is a workaround.
        asset = self._get_asset(AssetType.DATASET, name)
        if asset.status == AssetTrackerStatus.DONE:
            return

        self._update_asset(
            AssetType.DATASET,
            name,
            status=AssetTrackerStatus.BUILDING,
            url=self._get_url(AssetType.DATASET, name),
            version=version,
            build_idx=build_idx,
        )

    def mark_dataset_failed(self, name: str, reason: str) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            status=AssetTrackerStatus.ERROR,
            error_reason=f"{reason}",
        )

    def mark_dataset_built(
        self,
        name: str,
        *,
        version: Optional[str] = None,
        build_index: Optional[str] = None,
    ) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            status=AssetTrackerStatus.DONE,
            url=self._get_url(AssetType.DATASET, name),
            version=version,
            build_idx=build_index,
        )

    def mark_model_saving(self, name: str) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.SAVING,
        )

    def mark_model_saved(self, name: str) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.PENDING,
        )

    def mark_model_training(
        self, name: str, version: Optional[str] = None, train_idx: Optional[str] = None
    ) -> None:
        # For some reason, this function is called even after the model is successfully trained and
        # it causes a flicker in the progress bar. This is a workaround.
        asset = self._get_asset(AssetType.MODEL, name)
        if asset.status == AssetTrackerStatus.DONE:
            return

        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.TRAINING,
            url=self._get_url(AssetType.MODEL, name),
            version=version,
            build_idx=train_idx,
        )

    def mark_model_trained(
        self,
        name: str,
        *,
        version: Optional[str] = None,
        train_index: Optional[str] = None,
    ) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.DONE,
            url=self._get_url(AssetType.MODEL, name),
            version=version,
            build_idx=train_index,
        )

    def mark_model_train_failed(
        self, name: str, reason: str
    ) -> None:  # TODO(volkan) check that reason is always given
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.ERROR,
            error_reason=f"{reason}",
        )

    def update_dataset_saving_progress(
        self, name: str, cur_step: int, total_steps: int
    ) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            status=AssetTrackerStatus.SAVING,
            cur_step=cur_step,
            total_steps=total_steps,
        )

    def mark_dataset_resources_uploading(
        self, name: str, state: ResourceTransferState
    ) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            description="uploading",
            resource_transfer_state=state,
            status=AssetTrackerStatus.RESOURCE_UPLOADING,
        )

    def mark_dataset_resources_uploaded(self, name: str) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            description="uploaded",
            status=AssetTrackerStatus.PENDING,  # Pending here meaning that training has not started yet
        )

    def mark_model_resources_uploading(
        self, name: str, state: ResourceTransferState
    ) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            description="uploading",
            resource_transfer_state=state,
            status=AssetTrackerStatus.RESOURCE_UPLOADING,
        )

    def mark_model_resources_uploaded(self, name: str) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            description="uploaded",
            status=AssetTrackerStatus.PENDING,
        )

    def mark_model_running_assertions(self, name: str) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.ASSERTING,
        )

    def mark_model_running_assertion(self, name: str, assertion: Assertion) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.ASSERTING,
            description=str(assertion),
        )

    def mark_model_completed_assertions(self, name: str) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.ASSERTED,
        )

    def mark_model_failed_assertions(
        self, name: str, assertions: List[Assertion]
    ) -> None:
        stringified = [str(assertion) for assertion in assertions]
        error_msg = f"failed: {', '.join(stringified)}"
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.ERROR,
            error_reason=error_msg,
        )

    def mark_dataset_running_assertions(self, name: str) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            status=AssetTrackerStatus.ASSERTING,
        )

    def mark_dataset_running_assertion(self, name: str, assertion: Assertion) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            description=str(assertion),
        )

    def mark_dataset_completed_assertions(self, name: str) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            status=AssetTrackerStatus.ASSERTED,
        )

    def mark_dataset_failed_assertions(
        self, name: str, assertions: List[Assertion]
    ) -> None:
        stringified = [str(assertion) for assertion in assertions]
        error_msg = f"failed: {', '.join(stringified)}"
        self._update_asset(
            AssetType.DATASET,
            name,
            status=AssetTrackerStatus.ERROR,
            error_reason=error_msg,
        )

    def mark_dataset_saving_result(
        self, name: str, state: DatasetTransferState
    ) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            dataset_transfer_state=state,
            status=AssetTrackerStatus.RESULT_UPLOADING,
        )

    def mark_model_saving_result(self, name: str, state: ResourceTransferState) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            model_transfer_state=state,
            status=AssetTrackerStatus.RESULT_UPLOADING,
        )

    def mark_model_getting_model(
        self,
        name: str,
        getting_asset_name: str,
        state: Optional[ResourceTransferState],
        from_cache: bool,
    ) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            asset_download_transfer_state=state,
            status=AssetTrackerStatus.ASSET_DOWNLOADING
            if not from_cache
            else AssetTrackerStatus.ASSET_FROM_CACHE,
            loading_cache_asset=None if not from_cache else getting_asset_name,
        )

    def mark_model_getting_dataset(
        self, name: str, getting_asset_name: str, from_cache: bool
    ) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            asset_download_transfer_state=DatasetTransferState(0, getting_asset_name),
            status=AssetTrackerStatus.ASSET_DOWNLOADING
            if not from_cache
            else AssetTrackerStatus.ASSET_FROM_CACHE,
            loading_cache_asset=None if not from_cache else getting_asset_name,
        )

    def mark_dataset_getting_model(
        self,
        name: str,
        getting_asset_name: str,
        state: Optional[ResourceTransferState],
        from_cache: bool,
    ) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            asset_download_transfer_state=state,
            status=AssetTrackerStatus.ASSET_DOWNLOADING
            if not from_cache
            else AssetTrackerStatus.ASSET_FROM_CACHE,
            loading_cache_asset=None if not from_cache else getting_asset_name,
        )

    def mark_dataset_getting_dataset(
        self, name: str, getting_asset_name: str, from_cache: bool
    ) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            asset_download_transfer_state=DatasetTransferState(0, getting_asset_name),
            status=AssetTrackerStatus.ASSET_DOWNLOADING
            if not from_cache
            else AssetTrackerStatus.ASSET_FROM_CACHE,
            loading_cache_asset=None if not from_cache else getting_asset_name,
        )

    def mark_model_loaded(
        self,
        name: str,
    ) -> None:
        self._update_asset(
            AssetType.MODEL,
            name,
            status=AssetTrackerStatus.ASSET_LOADED,
        )

    def mark_dataset_loaded(
        self,
        name: str,
    ) -> None:
        self._update_asset(
            AssetType.DATASET,
            name,
            status=AssetTrackerStatus.ASSET_LOADED,
        )



import os
from logging import Logger
from pathlib import Path
from typing import Any, List, Optional, Union

from layer import global_context
from layer.clients.layer import LayerClient
from layer.config import ConfigManager
from layer.contracts.fabrics import Fabric
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.projects import Project, ProjectLoader
from layer.global_context import (
    set_default_fabric,
    set_pip_packages,
    set_pip_requirements_file,
)
from layer.projects.utils import get_or_create_remote_project
from layer.utils.async_utils import asyncio_run_in_thread


class InitProjectRunner:
    """Class responsible for setting up a user's project locally
    and creating a project in the Layer Backend if it doesn't exist already
    """

    def __init__(
        self,
        project_full_name: ProjectFullName,
        project_root_path: Optional[Union[str, Path]] = None,
        logger: Optional[Logger] = None,
        config_manager: Optional[ConfigManager] = None,
    ):
        self._project_full_name = project_full_name
        self._project_root_path = project_root_path
        self._config_manager = (
            config_manager if config_manager is not None else ConfigManager()
        )
        self._logger = (
            logger if logger is not None else Logger(__name__)  # not sure about this
        )

    def _ensure_user_logged_in(self) -> Any:
        login_config = asyncio_run_in_thread(self._config_manager.refresh())
        return login_config

    def _update_readme(
        self, project_full_name: ProjectFullName, layer_client: LayerClient
    ) -> None:
        readme_discover_path = self._project_root_path
        if not readme_discover_path:
            # We expect README file in the running directory unless told otherwise
            readme_discover_path = "./"
        project_root_path = Path(os.path.relpath(readme_discover_path))

        readme_contents = ProjectLoader.load_project_readme(project_root_path)
        if readme_contents:
            layer_client.project_service_client.update_project_readme(
                project_full_name=project_full_name,
                readme=readme_contents,
            )

    def setup_project(
        self,
        layer_client: Optional[LayerClient] = None,
        fabric: Optional[Fabric] = None,
        pip_packages: Optional[List[str]] = None,
        pip_requirements_file: Optional[str] = None,
    ) -> Project:
        if not layer_client:
            login_config = self._ensure_user_logged_in()
            layer_client = LayerClient(login_config.client, self._logger)

        with layer_client.init() as initialized_client:
            project = get_or_create_remote_project(
                initialized_client, self._project_full_name
            )

            self._update_readme(self._project_full_name, layer_client)

        global_context.reset_to(self._project_full_name)
        if fabric:
            set_default_fabric(fabric)
        if pip_packages:
            set_pip_packages(pip_packages)
        if pip_requirements_file:
            set_pip_requirements_file(pip_requirements_file)

        return project



from typing import Any, Callable, Dict, List

import numpy
import wrapt  # type: ignore

from layer.contracts.assertions import Assertion
from layer.decorators.layer_wrapper import LayerFunctionWrapper


class LayerAssertFunctionWrapper(LayerFunctionWrapper):
    def __init__(
        self,
        wrapped: Any,
        wrapper: Any,
        enabled: Any,
        assert_func: Callable[..., Any],
        values: List[Any],
    ) -> None:
        super().__init__(wrapped, wrapper, enabled)
        self.layer.append_assertion(
            Assertion(
                name=assert_func.__name__.lstrip("_"),
                values=values,
                function=assert_func(*values),
            )
        )


def assert_true(assert_function: Callable[..., bool]) -> Callable[..., Any]:
    """
    Asserts that a condition is true.
    This assertion can be used with Layer dataset and model entities.

    :param assert_function: Test function with a boolean return type.
    :return: Function object.

    .. code-block:: python

        from layer.decorators.assertions import assert_true
        from layer.decorators import dataset

        def test_function(data_frame):
            return len(data_frame.index) == 0

        @dataset("empty-data")
        @assert_true(test_function)
        def create_my_dataset():
            return pandas.DataFrame()
    """

    @wrapt.decorator(proxy=_assert_true_wrapper(assert_function))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _assert_true_wrapper(assert_function: Callable[..., bool]) -> Any:
    class FunctionWrapper(LayerAssertFunctionWrapper):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any = None) -> None:
            super().__init__(
                wrapped,
                wrapper,
                enabled,
                _assert_true,
                [assert_function],
            )

    return FunctionWrapper


def _assert_true(assert_function: Callable[..., Any]) -> Callable[[Any], Any]:
    def assert_func(to_check: Any) -> Any:
        assertion_result = assert_function(to_check)
        if not isinstance(assertion_result, (bool, numpy.bool_)):
            raise AssertionError(
                "Test FAILED: assert_true only accepts functions with boolean return type."
            )
        if assertion_result:
            print(f"Test SUCCESS: assert_true({assert_function.__name__}).")
        else:
            raise AssertionError(
                f"Test FAILED: assert_true({assert_function.__name__})."
            )
        return to_check

    return assert_func


def assert_valid_values(
    column_name: str, valid_values: List[Any]
) -> Callable[..., Any]:
    """
    Asserts that only the given valid values present in the given column.
    This assertion can be used with Layer dataset entities.

    :param column_name: Column name to be checked from the given dataframe.
    :param valid_values: Valid values.
    :return: Function object.

    .. code-block:: python

        from layer.decorators.assertions import assert_valid_values
        from layer.decorators import dataset

        @dataset("user-os-data")
        @assert_valid_values("OperatingSystem", ["linux", "other"])
        def create_my_dataset():
            data = [["user1", "linux"], ["user2", "linux"], ["user3", "other"]]
            dataframe = pd.DataFrame(data, columns=["UserName", "OperatingSystem"])
            return dataframe
    """

    @wrapt.decorator(proxy=_assert_valid_values_wrapper(column_name, valid_values))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _assert_valid_values_wrapper(column_name: str, valid_values: List[Any]) -> Any:
    if not isinstance(valid_values, list):
        raise AssertionError(
            "Test FAILED: assert_valid_values only accepts list type valid_values."
        )

    class FunctionWrapper(LayerAssertFunctionWrapper):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any = None) -> None:
            super().__init__(
                wrapped,
                wrapper,
                enabled,
                _assert_valid_values,
                [column_name, valid_values],
            )

    return FunctionWrapper


def _assert_valid_values(
    column_name: str, valid_values: List[Any]
) -> Callable[[Any], Any]:
    def assert_func(result: Any) -> Any:
        existing_unique_values = result[column_name].unique()

        invalid_values = list(set(existing_unique_values) - set(valid_values))
        if not invalid_values:
            print(
                f"Test SUCCESS: assert_valid_values('{column_name}', {valid_values})."
            )
        else:
            raise AssertionError(
                f"Test FAILED: assert_valid_values('{column_name}', {valid_values}). "
                f"Values {invalid_values} not in valid values."
            )

        return result

    return assert_func


def assert_not_null(column_names: List[str]) -> Callable[..., Any]:
    """
    Asserts that given columns do not have any null values.
    This assertion can be used with Layer dataset entities.

    :param column_names: List of columns to be checked.
    :return: Function object.

    .. code-block:: python

        from layer.decorators.assertions import assert_not_null
        from layer.decorators import dataset

        @dataset("user-os-data")
        @assert_not_null(["UserName", "OperatingSystem"])
        def create_my_dataset():
            # No null value in the dataframe. Prints a success message.
            data = [["user1", "linux"], ["user2", "linux"], ["user3", "other"]]
            dataframe = pd.DataFrame(data, columns=["UserName", "OperatingSystem"])
            return dataframe
    """

    @wrapt.decorator(proxy=_assert_not_null_wrapper(column_names))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _assert_not_null_wrapper(column_names: List[str]) -> Any:
    if not isinstance(column_names, list) or not all(
        isinstance(x, str) for x in column_names
    ):
        raise AssertionError(
            "Test FAILED: assert_not_null only accepts string list type column_names."
        )

    class FunctionWrapper(LayerAssertFunctionWrapper):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any = None) -> None:
            super().__init__(
                wrapped,
                wrapper,
                enabled,
                _assert_not_null,
                [column_names],
            )

    return FunctionWrapper


def _assert_not_null(column_names: List[str]) -> Callable[[Any], Any]:
    def assert_func(result: Any) -> Any:
        selected_test_data = result[column_names]
        columns_with_null = selected_test_data.columns[
            selected_test_data.isnull().any()
        ].tolist()

        if not columns_with_null:
            print(f"Test SUCCESS: assert_not_null({column_names}).")
        else:
            raise AssertionError(
                f"Test FAILED: assert_not_null({column_names}). "
                f"Columns {columns_with_null} have null values."
            )
        return result

    return assert_func


def assert_unique(column_subset: List[str]) -> Callable[..., Any]:
    """
    Asserts that given subset of columns together have no duplicate rows.
    This assertion can be used with Layer dataset entities.

    :param column_subset: Subset of the columns to be checked.
    :return: Function object.

    .. code-block:: python

        from layer.decorators.assertions import assert_unique
        from layer.decorators import dataset

        @dataset("user-data")
        @assert_unique(["Id", "UserName", "Company"])
        def create_my_dataset():
            # Row 1 and Row 2 are duplicates. It throws an error.
            data = [[1, "user1", "company1"], [1, "user1", "company1"], [1, "user1", "company2"]]
            dataframe = pd.DataFrame(data, columns=["Id", "UserName", "Company"])
            return dataframe
    """

    @wrapt.decorator(proxy=_assert_unique_wrapper(column_subset))
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _assert_unique_wrapper(column_subset: List[str]) -> Any:
    if not isinstance(column_subset, list) or not all(
        isinstance(x, str) for x in column_subset
    ):
        raise AssertionError(
            "Test FAILED: assert_unique only accepts string list type column_subset."
        )

    class FunctionWrapper(LayerAssertFunctionWrapper):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any = None) -> None:
            super().__init__(
                wrapped,
                wrapper,
                enabled,
                _assert_unique,
                [column_subset],
            )

    return FunctionWrapper


def _assert_unique(column_subset: List[str]) -> Callable[[Any], Any]:
    def assert_func(result: Any) -> Any:
        if not result.duplicated(subset=column_subset).any():
            print(f"Test SUCCESS: assert_unique({column_subset}).")
        else:
            raise AssertionError(
                f"Test FAILED: assert_unique({column_subset}). "
                f"Columns {column_subset} have duplicates."
            )
        return result

    return assert_func


def assert_skewness(
    column_name: str, min_skewness: float, max_skewness: float
) -> Callable[..., Any]:
    """
    Asserts that skewness value of the given column is in between the given minimum and maximum values.
    This assertion can be used with Layer dataset entities.

    :param column_name: Column to be checked.
    :param min_skewness: Accepted minimum skewness value.
    :param max_skewness: Accepted maximum skewness value.
    :return: Function object.

    .. code-block:: python

        from layer.decorators.assertions import assert_skewness
        from layer.decorators import dataset

        @dataset("product-data")
        @assert_skewness("Price", -0.3, 0.3)
        def create_my_dataset():
            # Assert success as skewness is lower than the threshold.
            data = [[1, "product1", 15], [2, "product2", 20], [3, "product3", 10]]
            dataframe = pd.DataFrame(data, columns=["Id", "Product", "Price"])
            return dataframe
    """

    @wrapt.decorator(
        proxy=_assert_skewness_wrapper(column_name, min_skewness, max_skewness)
    )
    def wrapper(
        wrapped: Any, instance: Any, args: List[Any], kwargs: Dict[str, Any]
    ) -> None:
        return wrapped(*args, **kwargs)

    return wrapper


def _assert_skewness_wrapper(
    column_name: str, min_skewness: float, max_skewness: float
) -> Any:
    if (
        not isinstance(column_name, str)
        or not isinstance(min_skewness, (int, float))
        or not isinstance(max_skewness, (int, float))
    ):
        raise AssertionError(
            "assert_skewness only accepts string type column_name and "
            "numeric type min_skewness and max_skewness values."
        )

    class FunctionWrapper(LayerAssertFunctionWrapper):
        def __init__(self, wrapped: Any, wrapper: Any, enabled: Any = None) -> None:
            super().__init__(
                wrapped,
                wrapper,
                enabled,
                _assert_skewness,
                [column_name, min_skewness, max_skewness],
            )

    return FunctionWrapper


def _assert_skewness(
    column_name: str, min_skewness: float, max_skewness: float
) -> Callable[[Any], Any]:
    def assert_func(result: Any) -> Any:
        skewness_value = result[column_name].skew(axis=0)
        in_range = max_skewness > skewness_value > min_skewness

        if in_range:
            print(
                f"Test SUCCESS: assert_skewness('{column_name}', {min_skewness}, {max_skewness})."
            )
        else:
            raise AssertionError(
                f"Test FAILED: assert_skewness('{column_name}', {min_skewness}, {max_skewness}). "
                f"Skewness value [{skewness_value}] of {column_name} column is not in the given range."
            )
        return result

    return assert_func



import logging
import threading
import uuid
from datetime import datetime
from typing import Any, Callable, List, Optional, Sequence

import polling  # type: ignore
from layerapi.api.entity.operations_pb2 import ExecutionPlan
from layerapi.api.ids_pb2 import RunId

from layer.clients.layer import LayerClient
from layer.config import Config
from layer.contracts.assets import AssetType
from layer.contracts.project_full_name import ProjectFullName
from layer.contracts.projects import ApplyResult
from layer.contracts.runs import FunctionDefinition, Run
from layer.exceptions.exceptions import (
    LayerClientException,
    LayerClientServiceUnavailableException,
    LayerClientTimeoutException,
    LayerResourceExhaustedException,
    LayerServiceUnavailableExceptionDuringExecution,
    LayerServiceUnavailableExceptionDuringInitialization,
    ProjectBaseException,
    ProjectInitializationException,
    ProjectRunnerError,
)
from layer.projects.execution_planner import (
    build_execution_plan,
    check_asset_dependencies,
)
from layer.projects.progress_tracker_updater import (
    PollingStepFunction,
    ProgressTrackerUpdater,
)
from layer.projects.utils import (
    calculate_hash_by_definitions,
    get_or_create_remote_project,
    verify_project_exists_and_retrieve_project_id,
)
from layer.resource_manager import ResourceManager
from layer.tracker.progress_tracker import RunProgressTracker
from layer.user_logs import LOGS_BUFFER_INTERVAL, show_pipeline_run_logs


logger = logging.getLogger()


class RunContext:
    def __init__(
        self, is_running: bool, run_completion_time: Optional[datetime] = None
    ) -> None:
        self._is_running = is_running
        self._run_completion_time: Optional[datetime] = run_completion_time

    @property
    def is_running(self) -> bool:
        return self._is_running

    @is_running.setter
    def is_running(self, is_running: bool) -> None:
        self._is_running = is_running

    @property
    def run_completion_time(self) -> Optional[datetime]:
        return self._run_completion_time

    @run_completion_time.setter
    def run_completion_time(self, run_completion_time: Optional[datetime]) -> None:
        self._run_completion_time = run_completion_time

    def user_logs_check_predicate(self) -> bool:
        if self._is_running:
            return True
        assert self._run_completion_time
        return (
            datetime.now() - self._run_completion_time
        ).total_seconds() < LOGS_BUFFER_INTERVAL


class ProjectRunner:
    _tracker: RunProgressTracker

    def __init__(
        self,
        config: Config,
    ) -> None:
        self._config = config

    def get_tracker(self) -> RunProgressTracker:
        return self._tracker

    def _apply(self, client: LayerClient, run: Run) -> ApplyResult:
        for definition in run.definitions:
            if definition.asset_type == AssetType.DATASET:
                register_dataset_function(client, definition, False, self._tracker)
            if definition.asset_type == AssetType.MODEL:
                register_model_function(client, definition, False, self._tracker)

        execution_plan = build_execution_plan(run)
        client.project_service_client.update_project_readme(
            run.project_full_name, run.readme
        )
        return ApplyResult(execution_plan=execution_plan)

    @staticmethod
    def with_functions(project_full_name: ProjectFullName, functions: List[Any]) -> Run:
        definitions: List[FunctionDefinition] = [f.get_definition() for f in functions]

        return Run(
            project_full_name=project_full_name,
            definitions=definitions,
            files_hash=calculate_hash_by_definitions(definitions),
        )

    def run(
        self,
        run: Run,
        debug: bool = False,
        printer: Callable[[str], Any] = print,
    ) -> Run:
        check_asset_dependencies(run.definitions)
        with LayerClient(self._config.client, logger).init() as client:
            get_or_create_remote_project(client, run.project_full_name)
            with RunProgressTracker(
                url=self._config.url,
                account_name=run.project_full_name.account_name,
                project_name=run.project_full_name.project_name,
                assets=[(d.asset_type, d.asset_name) for d in run.definitions],
            ).track() as tracker:
                self._tracker = tracker
                try:
                    metadata = self._apply(client, run)
                    ResourceManager(client).wait_resource_upload(run, tracker)
                    user_command = self._get_user_command(
                        execute_function=ProjectRunner.run, functions=run.definitions
                    )
                    run_id = self._run(
                        client, run, metadata.execution_plan, user_command
                    )
                    run = run.with_run_id(run_id)
                except LayerClientServiceUnavailableException as e:
                    raise LayerServiceUnavailableExceptionDuringInitialization(str(e))
                try:
                    if debug:
                        run_context = RunContext(is_running=True)
                        logs_thread = threading.Thread(
                            target=show_pipeline_run_logs,
                            args=(client, run_id.value, True),
                            kwargs={
                                "printer": printer,
                                "evaluate_until_callback": run_context.user_logs_check_predicate,
                            },
                        )
                        logs_thread.start()
                    self._tracker.mark_start_running(run_id)
                    try:
                        self._poll_until_completed(client, metadata, run)
                    except (ProjectBaseException, ProjectRunnerError) as e:
                        tracker.mark_error_messages(e)
                        raise e
                    finally:
                        if debug:
                            self._finish_checking_user_logs(run_context, logs_thread)
                except LayerClientServiceUnavailableException as e:
                    raise LayerServiceUnavailableExceptionDuringExecution(
                        run_id, str(e)
                    )

        return run

    @staticmethod
    def _get_user_command(
        execute_function: Callable[..., Any], functions: Sequence[FunctionDefinition]
    ) -> str:
        functions_string = ", ".join(function.func_name for function in functions)
        return f"{execute_function.__name__}([{functions_string}])"

    @staticmethod
    def _finish_checking_user_logs(
        run_context: RunContext, logs_thread: threading.Thread
    ) -> None:
        run_context.run_completion_time = datetime.now()
        run_context.is_running = False
        logs_thread.join()

    def _poll_until_completed(
        self, client: LayerClient, apply_metadata: ApplyResult, run: Run
    ) -> None:
        updater = ProgressTrackerUpdater(
            tracker=self._tracker,
            apply_metadata=apply_metadata,
            run=run,
            client=client,
        )

        initial_step_sec = 2
        step_function = PollingStepFunction(max_backoff_sec=3.0, backoff_multiplier=1.2)

        assert run.run_id
        polling.poll(
            lambda: client.flow_manager.get_run_status_history_and_metadata(
                run_id=run.run_id,
            ),
            check_success=updater.check_completion_and_update_tracker,
            step=initial_step_sec,
            step_function=step_function.step,
            ignore_exceptions=(LayerClientTimeoutException,),
            poll_forever=True,
        )

    @staticmethod
    def _run(
        client: LayerClient,
        run: Run,
        execution_plan: ExecutionPlan,
        user_command: str,
    ) -> RunId:
        try:
            run_id = client.flow_manager.start_run(
                run.project_full_name, execution_plan, run.files_hash, user_command
            )
        except LayerResourceExhaustedException as e:
            raise ProjectRunnerError(f"{e}")
        else:
            return run_id

    def terminate_run(self, run_id: str) -> None:
        _run_id = RunId(value=run_id)
        with LayerClient(self._config.client, logger).init() as client:
            client.flow_manager.terminate_run(_run_id)


def register_dataset_function(
    client: LayerClient,
    dataset: FunctionDefinition,
    is_local: bool,
    tracker: RunProgressTracker,
) -> None:
    try:
        project_id = verify_project_exists_and_retrieve_project_id(
            client, dataset.project_full_name
        )
        dataset_id = client.data_catalog.add_dataset(project_id, dataset, is_local)
        dataset.set_repository_id(uuid.UUID(dataset_id))
        assert dataset.repository_id
        tracker.mark_dataset_saved(dataset.asset_name, id_=dataset.repository_id)
    except LayerClientServiceUnavailableException as e:
        tracker.mark_dataset_failed(dataset.asset_name, "")
        raise LayerServiceUnavailableExceptionDuringInitialization(str(e))
    except LayerClientException as e:
        tracker.mark_dataset_failed(dataset.asset_name, "")
        raise ProjectInitializationException(
            f"Failed to save derived dataset {dataset.asset_name!r}: {e}",
            "Please retry",
        )


def register_model_function(
    client: LayerClient,
    model: FunctionDefinition,
    is_local: bool,
    tracker: RunProgressTracker,
) -> None:
    try:
        response = client.model_catalog.create_model_version(
            model.project_full_name, model, is_local
        )
        version = response.model_version
        model.set_version_id(version.id.value)
        if response.should_upload_training_files:
            # in here we upload to path / train.gz
            version_id = uuid.UUID(version.id.value)
            s3_path = client.model_training.upload_training_files(model, version_id)
            # in here we reconstruct the path / train.gz to save in metadata
            client.model_catalog.store_training_metadata(model, s3_path, version)
        tracker.mark_model_saved(model.asset_name)
    except LayerClientServiceUnavailableException as e:
        tracker.mark_model_train_failed(model.asset_name, "")
        raise LayerServiceUnavailableExceptionDuringInitialization(str(e))
    except LayerClientException as e:
        tracker.mark_model_train_failed(model.asset_name, "")
        raise ProjectInitializationException(
            f"Failed to save model {model.asset_name!r}: {e}",
            "Please retry",
        )



from typing import TYPE_CHECKING, Any, Dict, Optional

from yarl import URL

from layer.config import (
    AccountServiceConfig,
    AuthConfig,
    ClientConfig,
    Config,
    DataCatalogConfig,
    FlowManagerServiceConfig,
    ModelCatalogConfig,
    ModelTrainingConfig,
    ProjectServiceConfig,
    S3Config,
    UserLogsServiceConfig,
)


if TYPE_CHECKING:
    from aiohttp import ClientSession


class ConfigClient:
    def __init__(
        self, *, url: URL, client: "ClientSession", do_verify_ssl: bool = True
    ) -> None:
        self._url = url
        self._details_url = url / "__details.json"
        self._client = client
        self._do_verify_ssl = do_verify_ssl
        self._client_ssl_param = None if do_verify_ssl else False

    async def get_config(self) -> Config:
        async with self._client.get(
            self._details_url, ssl=self._client_ssl_param
        ) as resp:
            payload = await resp.json()
            return Config(
                url=self._url,
                auth=self._create_auth_config(payload.get("auth", {})),
                client=self._create_client_config(payload.get("client", {})),
            )

    def _create_auth_config(
        self, payload: Dict[str, Any], url: Optional[URL] = None
    ) -> AuthConfig:
        if not payload:
            return AuthConfig.create_disabled()
        if not url:
            url = self._url
        base_url = URL.build(scheme="https", host=payload["domain"])
        sdk_client_payload = payload["clients"]["sdk"]
        callback_urls = [
            URL(callback_url)
            for callback_url in sdk_client_payload["callback_urls"]
            if "127.0.0.1" in callback_url
        ]
        return AuthConfig(
            auth_url=url / "oauth" / "authorize",
            token_url=base_url / "oauth" / "token",
            logout_url=base_url / "v2" / "logout",
            client_id=sdk_client_payload["client_id"],
            # Our 'audience' of Auth0 OAuth config should not have a trailing slash, otherwise it fails.
            audience=str(self._remove_trailing_slash(url)),
            headless_callback_url=url / "oauth" / "code",
            callback_urls=callback_urls,
            success_redirect_url=url,
            failure_redirect_url=url / "oauth" / "code",
        )

    @staticmethod
    def _remove_trailing_slash(url: URL) -> URL:
        return URL(str(url).rstrip("/"))

    def _create_client_config(self, payload: Dict[str, Any]) -> ClientConfig:
        if "grpc_gateway_url" in payload:
            url = URL(payload["grpc_gateway_url"])
            do_verify_ssl = payload.get("grpc_do_verify_ssl", True)
        else:
            url = self._url.with_host(f"grpc.{self._url.host}")
            do_verify_ssl = self._do_verify_ssl
        grpc_gateway_address = f"{url.host}:{url.port}"
        return ClientConfig(
            data_catalog=DataCatalogConfig(address=grpc_gateway_address),
            model_catalog=ModelCatalogConfig(address=grpc_gateway_address),
            model_training=ModelTrainingConfig(address=grpc_gateway_address),
            account_service=AccountServiceConfig(address=grpc_gateway_address),
            flow_manager=FlowManagerServiceConfig(address=grpc_gateway_address),
            user_logs=UserLogsServiceConfig(address=grpc_gateway_address),
            project_service=ProjectServiceConfig(address=grpc_gateway_address),
            grpc_gateway_address=grpc_gateway_address,
            grpc_do_verify_ssl=do_verify_ssl,
            s3=self._create_s3_config(payload),
        )

    def _create_s3_config(self, payload: Dict[str, Any]) -> S3Config:
        s3_endpoint_url = None
        if payload.get("s3_endpoint_url"):
            s3_endpoint_url = URL(payload["s3_endpoint_url"])
        return S3Config(endpoint_url=s3_endpoint_url)



import time
import uuid
from typing import Any, Optional
from uuid import UUID

from layerapi.api.ids_pb2 import ModelTrainId
from layerapi.api.value.model_flavor_pb2 import ModelFlavor

from layer.clients.layer import LayerClient
from layer.contracts.models import Model
from layer.exceptions.exceptions import UnexpectedModelTypeException
from layer.flavors.utils import get_flavor_for_model
from layer.tracker.progress_tracker import RunProgressTracker
from layer.types import ModelObject

from ..contracts.project_full_name import ProjectFullName
from .base_train import BaseTrain


class Train(BaseTrain):
    """
    Allows the user to control the state of a model training.

    The Train object methods can then be called to log parameters or metrics during training.

    This class should never be instantiated directly by end users.

    .. code-block:: python

        def train_model(train: Train, tf: Featureset("transaction_features")):
            # We create the training and label data
            train_df = tf.to_pandas()
            X = train_df.drop(["transactionId", "is_fraud"], axis=1)
            Y = train_df["is_fraud"]
            # user model training will happen below [...]
    """

    def __init__(
        self,
        layer_client: LayerClient,
        name: str,
        project_full_name: ProjectFullName,
        version: Optional[str],
        train_id: Optional[UUID] = None,
        train_index: Optional[str] = None,
    ):
        self.__layer_client: LayerClient = layer_client
        self.__name: str = name
        self.__project_full_name: ProjectFullName = project_full_name
        self.__version: Optional[str] = str(version)
        self.__train_index: Optional[str] = str(train_index)
        self.__train_id: Optional[ModelTrainId] = (
            ModelTrainId(value=str(train_id)) if train_id is not None else None
        )
        self.__start_train_ts: int  # For computing relative to start metric timestamps
        # Populated at the save of a model train
        self.__flavor: Optional[ModelFlavor.V] = None

    def get_id(self) -> UUID:
        assert self.__train_id
        return UUID(self.__train_id.value)

    def get_version(self) -> str:
        assert self.__version
        return self.__version

    def get_train_index(self) -> str:
        assert self.__train_index
        return self.__train_index

    def save_model(
        self,
        model_object: ModelObject,
        tracker: RunProgressTracker,
    ) -> Any:
        assert self.__train_id

        flavor = get_flavor_for_model(model_object)
        if flavor is None:
            raise UnexpectedModelTypeException(type(model_object))
        storage_config = (
            self.__layer_client.model_catalog.get_model_train_storage_configuration(
                self.__train_id
            )
        )
        self.__flavor = flavor.PROTO_FLAVOR
        model = Model(
            self.__name,
            uuid.UUID(self.__train_id.value),
            flavor=flavor,
            storage_config=storage_config,
        )
        self.__layer_client.model_catalog.save_model_object(
            model, model_object, tracker=tracker
        )

    def __start_train(self) -> None:
        if self.__train_id is None:
            self.__train_id = self.__layer_client.model_catalog.create_model_train(
                name=self.__name,
                version=self.__version,
                project_full_name=self.__project_full_name,
            )
        self.__layer_client.model_catalog.start_model_train(
            train_id=self.__train_id,
        )
        self.__start_train_ts = int(time.time())

    def __complete_train(self) -> None:
        assert self.__train_id
        self.__layer_client.model_catalog.complete_model_train(
            self.__train_id, self.__flavor
        )

    def __enter__(self) -> Any:
        self.__start_train()
        return self

    def __exit__(
        self, exception_type: Any, exception_value: Any, traceback: Any
    ) -> None:
        # From the docs we don't need to re-raise the exception as it will be thrown after
        # the execution of this method. We'd only proceed if there are no exceptions.
        # In happy path, all three parameters would be None.
        # https://docs.python.org/3/reference/datamodel.html#object.__exit__
        if exception_type is None:
            self.__complete_train()



import abc
import uuid
from dataclasses import dataclass, field
from typing import Optional
from uuid import UUID


@dataclass(frozen=True)
class Account:
    id: UUID
    name: str


@dataclass(frozen=True)
class User(abc.ABC):
    name: str
    email: str
    first_name: str
    last_name: str
    id: uuid.UUID = field(default_factory=uuid.uuid4)
    account_id: Optional[uuid.UUID] = field(default_factory=uuid.uuid4)



import logging
import uuid
import warnings
from collections import namedtuple
from dataclasses import dataclass, replace
from pathlib import Path
from typing import Optional

from layerapi.api.entity.operations_pb2 import ExecutionPlan

from layer.exceptions.exceptions import ProjectException

from .accounts import Account
from .project_full_name import ProjectFullName


logger = logging.getLogger()
LeveledNode = namedtuple("LeveledNode", ["node", "level"])


@dataclass(frozen=True)
class Project:
    """
    Provides access to projects stored in Layer. Projects are containers to organize your machine learning project entities.

    You can retrieve an instance of this object with :code:`layer.init()`.

    This class should not be initialized by end-users.

    .. code-block:: python

        # Initializes a project with name "my-project"
        layer.init("my-project")

    """

    name: str
    id: uuid.UUID
    account: Account

    @property
    def full_name(self) -> ProjectFullName:
        if self.account is None:
            raise ProjectException("project has no account defined")
        return ProjectFullName(
            project_name=self.name,
            account_name=self.account.name,
        )

    def with_name(self, name: str) -> "Project":
        """
        :return: A new object that has a new name but all other fields are the same.
        """
        return replace(self, name=name)

    def with_id(self, project_id: uuid.UUID) -> "Project":
        """
        :return: A new object that has a new id but all other fields are the same.
        """
        return replace(self, id=project_id)

    def with_account(self, account: Account) -> "Project":
        return replace(self, account=account)

    def __str__(self) -> str:
        if self.account:
            from layer.config import DEFAULT_PATH, ConfigManager

            config = ConfigManager(DEFAULT_PATH).load()
            return f"Your Layer project is here: {config.url}/{self.account.name}/{self.name}"
        else:
            return f"Project({self.name})"

    def __repr__(self) -> str:
        return self.__str__()


@dataclass(frozen=True)
class ApplyResult:
    execution_plan: ExecutionPlan


class ProjectLoader:
    @staticmethod
    def load_project_readme(path: Path) -> Optional[str]:
        for subp in path.glob("*"):
            if subp.name.lower() == "readme.md":
                with open(subp, "r") as f:
                    readme = f.read()
                    # restrict length of text we send to backend
                    if len(readme) > 25_000:
                        warnings.warn(
                            "Your README.md will be truncated to 25000 characters",
                        )
                        readme = readme[:25_000]
                    return readme
        return None
